STDIN = "stdin"


class VTracker:
    """This helper class is used to identify the unique variants from the input
    so the crv file will not contain multiple copies of the same variant.
    """

    def __init__(self, deduplicate=True):
        from collections import defaultdict

        self.var_by_chrom = defaultdict(dict)
        self.current_UID = 1
        self.deduplicate = deduplicate

    # Add a variant - Returns true if the variant is a new unique variant, false
    # if it is a duplicate.  Also returns the UID.
    def addVar(self, chrom, pos, ref, alt):
        if not self.deduplicate:
            self.current_UID += 1
            return True, self.current_UID - 1

        change = ref + ":" + alt

        chr_dict = self.var_by_chrom[chrom]
        if pos not in chr_dict:
            # we have not seen this position before, add the position and change
            chr_dict[pos] = {}
            chr_dict[pos][change] = self.current_UID
            self.current_UID += 1
            return True, chr_dict[pos][change]
        else:
            variants = chr_dict[pos]
            if change not in variants:
                # we have the position but not this base change, add it.
                chr_dict[pos][change] = self.current_UID
                self.current_UID = self.current_UID + 1
                return True, chr_dict[pos][change]
            else:
                # this variant has been seen before.
                return False, chr_dict[pos][change]


class MasterCravatConverter(object):
    """Convert a file of ambiguous format to .crv format.

    Reads in CravatConverter classes in the same directory, selects the
    correct converter, and writes a crv file.
    """

    ALREADYCRV = 2

    def __init__(self, *inargs, **inkwargs):
        from oakvar.consts import crs_def
        from oakvar import get_wgs_reader

        self.logger = None
        self.crv_writer = None
        self.crs_writer = None
        self.crm_writer = None
        self.crl_writer = None
        self.primary_converter = None
        self.converters = {}
        self.possible_formats = []
        self.ready_to_convert = False
        self.input_format = None
        self.pipeinput = False
        self.input_paths = None
        self.input_dir = None
        self.input_path_dict = None
        self.input_path_dict2 = None
        self.output_dir = None
        self.output_base_fname = None
        self.status_fpath = None
        self.conf = None
        self.unique_variants = None
        self.status_writer = None
        self.error_logger = None
        self.unique_excs = []
        self.wpath = None
        self.err_path = None
        self.crm_path = None
        self.crs_path = None
        self.crl_path = None
        self.cur_file = None
        self.file_error_lines = 0
        self.total_error_lines = 0
        self.chromdict = {
            "chrx": "chrX",
            "chry": "chrY",
            "chrMT": "chrM",
            "chrMt": "chrM",
            "chr23": "chrX",
            "chr24": "chrY",
        }
        self._parse_cmd_args(inargs, inkwargs)
        self._setup_logger()
        self.vtracker = VTracker(deduplicate=not (self.unique_variants))
        self.wgsreader = get_wgs_reader(assembly="hg38")
        self.crs_def = crs_def.copy()

    def get_genome_assembly(self, converter):
        from oakvar.system.consts import default_assembly_key
        from oakvar.exceptions import NoGenomeException
        from oakvar.system import get_user_conf
    
        if self.given_input_assembly:
            return self.given_input_assembly
        input_assembly = getattr(converter, "input_assembly", None)
        if input_assembly:
            return input_assembly
        user_conf = get_user_conf() or {}
        genome_assembly = user_conf.get(default_assembly_key, None)
        if genome_assembly:
            return genome_assembly
        raise NoGenomeException()

    def get_do_liftover(self, genome_assembly):
        return genome_assembly != "hg38"

    def setup_lifter(self, genome_assembly):
        from oakvar.util.admin_util import get_liftover_chain_paths
        from oakvar.exceptions import InvalidGenomeAssembly
        from pyliftover import LiftOver
        lifter = None
        liftover_chain_paths = get_liftover_chain_paths()
        if genome_assembly not in liftover_chain_paths:
            raise InvalidGenomeAssembly(genome_assembly)
        lifter = LiftOver(liftover_chain_paths[genome_assembly])
        return lifter

    def _parse_cmd_args(self, inargs, inkwargs):
        """Parse the arguments in sys.argv"""
        import sys
        from json import loads
        from argparse import ArgumentParser, SUPPRESS
        from os.path import abspath, dirname, exists, basename, join
        from os import makedirs
        from oakvar.exceptions import ExpectedException
        from oakvar.util.util import get_args
        from oakvar.exceptions import SetupError

        parser = ArgumentParser()
        parser.add_argument("path", help="Path to this converter's python module")
        parser.add_argument(
            "inputs", nargs="*", default=None, help="Files to be converted to .crv"
        )
        parser.add_argument(
            "-f", dest="format", default=None, help="Specify an input format"
        )
        parser.add_argument(
            "-n", "--name", dest="name", help="Name of job. Default is input file name."
        )
        parser.add_argument(
            "-d",
            "--output-dir",
            dest="output_dir",
            help="Output directory. Default is input file directory.",
        )
        parser.add_argument(
            "-l",
            "--genome",
            dest="genome",
            default=None,
            help="Input gene assembly. Will be lifted over to hg38",
        )
        parser.add_argument(
            "--confs", dest="confs", default="{}", help="Configuration string"
        )
        parser.add_argument(
            "--unique-variants",
            dest="unique_variants",
            default=False,
            action="store_true",
            help=SUPPRESS,
        )
        if len(sys.argv) > 1 and len(inargs) == 0:
            inargs = [sys.argv]
        parsed_args = get_args(parser, inargs, inkwargs)
        self.input_format = None
        if parsed_args["format"]:
            self.input_format = parsed_args["format"]
        if parsed_args["inputs"] is None:
            raise ExpectedException("Input files are not given.")
        self.pipeinput = False
        if (
            parsed_args["inputs"] is not None
            and len(parsed_args["inputs"]) == 1
            and parsed_args["inputs"][0] == "-"
        ):
            self.pipeinput = True
        self.input_paths = []
        if self.pipeinput == False:
            self.input_paths = [abspath(x) for x in parsed_args["inputs"] if x != "-"]
        else:
            self.input_paths = [f"./{STDIN}"]
        self.input_dir = dirname(self.input_paths[0])
        self.input_path_dict = {}
        self.input_path_dict2 = {}
        if self.pipeinput == False:
            for i in range(len(self.input_paths)):
                self.input_path_dict[i] = self.input_paths[i]
                self.input_path_dict2[self.input_paths[i]] = i
        else:
            self.input_path_dict[0] = self.input_paths[0]
            self.input_path_dict2[STDIN] = 0
        self.output_dir = None
        if parsed_args["output_dir"]:
            self.output_dir = parsed_args["output_dir"]
        else:
            self.output_dir = self.input_dir
        if self.output_dir is None:
            raise SetupError("output directory")
        if not (exists(self.output_dir)):
            makedirs(self.output_dir)
        self.output_base_fname = None
        if parsed_args["name"]:
            self.output_base_fname = parsed_args["name"]
        else:
            self.output_base_fname = basename(self.input_paths[0])
        self.given_input_assembly = parsed_args["genome"]
        self.status_fpath = join(
            self.output_dir, self.output_base_fname + ".status.json"
        )
        self.conf = {}
        if parsed_args["confs"] is not None:
            confs = parsed_args["confs"].lstrip("'").rstrip("'").replace("'", '"')
            self.conf = loads(confs)
        if "conf" in parsed_args:
            self.conf.update(parsed_args["conf"])
        self.unique_variants = parsed_args["unique_variants"]
        if "status_writer" in parsed_args:
            self.status_writer = parsed_args["status_writer"]
        self.args = parsed_args

    def open_input_file(self, input_path):
        import gzip
        from oakvar.util.util import detect_encoding

        encoding = detect_encoding(input_path)
        if input_path.endswith(".gz"):
            f = gzip.open(input_path, mode="rt", encoding=encoding)
        else:
            f = open(input_path, encoding=encoding)
        return f

    def first_input_file(self):
        import gzip
        from sys import stdin
        from oakvar.util.util import detect_encoding
        from oakvar.exceptions import NoInput

        if self.input_paths is None:
            raise NoInput()
        if self.pipeinput == False:
            input_path = self.input_paths[0]
            encoding = detect_encoding(input_path)
            if input_path.endswith(".gz"):
                f = gzip.open(input_path, mode="rt", encoding=encoding)
            else:
                f = open(input_path, encoding=encoding)
        else:
            f = stdin
        return f

    def setup(self):
        """Do necesarry pre-run tasks"""
        if self.ready_to_convert:
            return
        # Read in the available converters
        self._initialize_converters()
        # Select the converter that matches the input format
        self._select_primary_converter()
        # Open the output files
        self._open_output_files()
        self.ready_to_convert = True

    def _setup_logger(self):
        """Open a log file and set up log handler"""
        from logging import getLogger
        from time import asctime

        self.logger = getLogger("oakvar.converter")
        self.logger.info("started: %s" % asctime())
        self.error_logger = getLogger("err.converter")
        self.unique_excs = []

    def _initialize_converters(self):
        """Reads in available converters.

        Loads any python files in same directory that start with _ as
        python modules. Initializes the CravatConverter class from that
        module and places them in a dict keyed by their input format
        """
        from oakvar.exceptions import ExpectedException
        from oakvar.module.local import get_local_module_infos_of_type
        from oakvar.exceptions import InvalidModule
        from oakvar.util.util import load_class

        for module_info in get_local_module_infos_of_type("converter").values():
            cls = load_class(module_info.script_path)
            converter = cls()
            converter.script_path = module_info.script_path
            if not hasattr(converter, "format_name"):
                raise InvalidModule(module_info.name)
            converter.module_name = module_info.name
            if converter.format_name not in self.converters:
                self.converters[converter.format_name] = converter
            else:
                err_msg = (
                    "Cannot load two converters for format %s" % converter.format_name
                )
                raise ExpectedException(err_msg)
        self.possible_formats = list(self.converters.keys())

    def _select_primary_converter(self):
        """Choose the converter which matches the input format.

        If a input format was not specified in the cmd args, uses the
        check_format() method of the CravatConverters to identify a
        converter which can parse the input file.
        """
        from oakvar.exceptions import InvalidInputFormat
        from oakvar.exceptions import LoggerError
        from oakvar.exceptions import NoInput
        from logging import getLogger

        if self.logger is None:
            raise LoggerError()
        if self.input_format is not None:
            if self.input_format not in self.possible_formats:
                raise InvalidInputFormat(self.input_format)
        else:
            if self.pipeinput == False:
                valid_formats = []
                first_file = self.first_input_file()
                first_file.seek(0)
                for converter_name, converter in self.converters.items():
                    try:
                        check_success = converter.check_format(first_file)
                    except:
                        check_success = False
                    first_file.seek(0)
                    if check_success:
                        valid_formats.append(converter_name)
                if len(valid_formats) == 0:
                    raise InvalidInputFormat(f"no converter for input found")
                elif len(valid_formats) > 1:
                    raise InvalidInputFormat(f"ambiguous {','.join(valid_formats)}")
                else:
                    self.input_format = valid_formats[0]
            else:
                if self.input_format is None:
                    raise InvalidInputFormat(
                        f"--input-format should be given with pipe input"
                    )
        self.primary_converter = self.converters[self.input_format]
        self._set_converter_properties(self.primary_converter)
        if self.input_paths is None:
            raise NoInput()
        if self.pipeinput == False:
            if len(self.input_paths) > 1:
                for fn in self.input_paths[1:]:
                    f = self.open_input_file(fn)
                    if not self.primary_converter.check_format(f):
                        raise InvalidInputFormat("inconsistent input formats")
        self.logger.info(
            f"version: {self.primary_converter.module_name}=={self.primary_converter.version} {self.primary_converter.script_path}"
        )
        self.error_logger = getLogger("err." + self.primary_converter.module_name)

    def _set_converter_properties(self, converter):
        from oakvar.exceptions import SetupError
        from oakvar.module.local import get_module_code_version

        if self.primary_converter is None or self.conf is None:
            raise SetupError()
        converter.output_dir = self.output_dir
        converter.run_name = self.output_base_fname
        module_name = self.primary_converter.format_name + "-converter"
        converter.module_name = module_name
        converter.version = get_module_code_version(converter.module_name)
        if module_name in self.conf:
            if hasattr(converter, "conf") == False:
                converter.conf = {}
            converter.conf.update(self.conf[module_name])

    def _open_output_files(self):
        from oakvar.exceptions import SetupError

        """Open .crv .crs and .crm output files, plus .err file.

        .crv .crs and .crm files are opened using a CravatWriter.
        .err file will contain all errors which occur during conversion.
        .map file contains two columns showing which lines in input
        correspond to which lines in output.
        """
        if (
            self.output_base_fname is None
            or self.primary_converter is None
            or self.output_dir is None
        ):
            raise SetupError()
        from oakvar.consts import (
            crv_def,
            crv_idx,
            crm_def,
            crm_idx,
            crs_idx,
            crl_def,
        )
        from oakvar.util.inout import CravatWriter
        from os.path import join

        # Setup CravatWriter
        self.wpath = join(self.output_dir, self.output_base_fname + ".crv")
        self.crv_writer = CravatWriter(self.wpath)
        self.crv_writer.add_columns(crv_def)
        self.crv_writer.write_definition()
        for index_columns in crv_idx:
            self.crv_writer.add_index(index_columns)
        self.crv_writer.wf.write(
            "#input_format={}\n".format(self.primary_converter.format_name)
        )
        # Setup err file
        self.err_path = join(self.output_dir, self.output_base_fname + ".converter.err")
        # Setup crm line mappings file
        self.crm_path = join(self.output_dir, self.output_base_fname + ".crm")
        self.crm_writer = CravatWriter(self.crm_path)
        self.crm_writer.add_columns(crm_def)
        self.crm_writer.write_definition()
        for index_columns in crm_idx:
            self.crm_writer.add_index(index_columns)
        self.crm_writer.write_input_paths(self.input_path_dict)
        # Setup crs sample file
        self.crs_path = join(self.output_dir, self.output_base_fname + ".crs")
        self.crs_writer = CravatWriter(self.crs_path)
        self.crs_writer.add_columns(self.crs_def)
        if hasattr(self.primary_converter, "addl_cols"):
            self.crs_writer.add_columns(self.primary_converter.addl_cols, append=True)
            self.crs_def.extend(self.primary_converter.addl_cols)
        self.crs_writer.write_definition()
        for index_columns in crs_idx:
            self.crs_writer.add_index(index_columns)
        self.crl_path = join(
            self.output_dir,
            ".".join([self.output_base_fname, "original_input", "var"]),
        )
        self.crl_writer = CravatWriter(self.crl_path)
        self.crl_writer.add_columns(crl_def)
        self.crl_writer.write_definition()
        self.crl_writer.write_names("original_input", "Original Input", "")

    def log_input_and_genome_assembly(self, input_path, genome_assembly):
        if not self.logger:
            return
        if self.pipeinput:
            self.logger.info(f"input file: {STDIN}")
        else:
            self.logger.info(f"input file: {input_path}")
        self.logger.info("input format: %s" % self.input_format)
        self.logger.info(f"genome_assembly: {genome_assembly}")

    def run(self):
        """Convert input file to a .crv file using the primary converter."""
        from sys import stdin
        from os.path import basename
        from time import time, asctime, localtime
        from copy import copy
        from re import compile
        from oakvar.exceptions import IgnoredVariant, NoVariantError
        from oakvar.base.converter import BaseConverter
        from oakvar.exceptions import SetupError
        from oakvar.exceptions import LoggerError
        from oakvar.exceptions import InvalidModule
        from oakvar.util.util import standardize_pos_ref_alt

        self.setup()
        if (
            self.wgsreader is None
            or self.crv_writer is None
            or self.crl_writer is None
            or self.crm_writer is None
            or self.crs_writer is None
            or self.input_path_dict2 is None
            or self.primary_converter is None
        ):
            raise SetupError()
        if self.logger is None:
            raise LoggerError()
        if not hasattr(self.primary_converter, "format_name"):
            if hasattr(self.primary_converter, "module_name"):
                raise InvalidModule(self.primary_converter.module_name)
            else:
                raise InvalidModule("unknown module")
        start_time = time()
        if self.status_writer is not None:
            self.status_writer.queue_status_update(
                "status",
                "Started {} ({})".format(
                    "Converter", self.primary_converter.format_name
                ),  # type: ignore
            )
        last_status_update_time = time()
        if self.input_paths is None:
            raise SetupError()
        multiple_files = len(self.input_paths) > 1
        fileno = 0
        total_lnum = 0
        base_re = compile("^[ATGC]+|[-]+$")
        write_lnum = 0
        genome_assemblies = set()
        for fn in self.input_paths:
            self.cur_file = fn
            if self.pipeinput:
                f = stdin
            else:
                f = self.open_input_file(fn)
            if self.pipeinput == True:
                fname = STDIN
            else:
                fname = f.name
            fileno += 1
            converter = self.primary_converter.__class__()
            if converter is None:
                raise SetupError()
            self._set_converter_properties(converter)
            converter.setup(f)  # type: ignore
            genome_assembly = self.get_genome_assembly(converter)
            genome_assemblies.add(genome_assembly)
            self.log_input_and_genome_assembly(fn, genome_assembly)
            do_liftover = self.get_do_liftover(genome_assembly)
            lifter = None
            if do_liftover:
                lifter = self.setup_lifter(genome_assembly)
            if self.pipeinput == False:
                f.seek(0)
            if self.pipeinput:
                cur_fname = STDIN
            else:
                cur_fname = basename(f.name)
            last_read_lnum = None
            self.file_error_lines = 0
            for read_lnum, l, all_wdicts in converter.convert_file(  # type: ignore
                f, exc_handler=self._log_conversion_error
            ):
                samp_prefix = cur_fname
                last_read_lnum = read_lnum
                try:
                    # all_wdicts is a list, since one input line can become
                    # multiple output lines. False is returned if converter
                    # decides line is not an input line.
                    if all_wdicts is BaseConverter.IGNORE:
                        continue
                    total_lnum += 1
                    if all_wdicts:
                        UIDMap = []
                        no_unique_var = 0
                        for wdict in all_wdicts:
                            chrom = wdict["chrom"]
                            pos = wdict["pos"]
                            if chrom is not None:
                                if not chrom.startswith("chr"):
                                    chrom = "chr" + chrom
                                wdict["chrom"] = self.chromdict.get(chrom, chrom)
                                if multiple_files:
                                    if wdict["sample_id"]:
                                        wdict["sample_id"] = "__".join(
                                            [samp_prefix, wdict["sample_id"]]
                                        )
                                    else:
                                        wdict["sample_id"] = samp_prefix
                                if "ref_base" not in wdict or wdict["ref_base"] == "":
                                    wdict["ref_base"] = self.wgsreader.get_bases(
                                        chrom, int(wdict["pos"])
                                    )
                                else:
                                    ref_base = wdict["ref_base"]
                                    if ref_base == "" and wdict["alt_base"] not in [
                                        "A",
                                        "T",
                                        "C",
                                        "G",
                                    ]:
                                        e = IgnoredVariant(
                                            "Reference base required for non SNV"
                                        )
                                        e.traceback = False
                                        raise e
                                    elif ref_base is None or ref_base == "":
                                        wdict["ref_base"] = self.wgsreader.get_bases(
                                            chrom, int(pos)
                                        )
                                prelift_wdict = copy(wdict)
                                if do_liftover:
                                    (
                                        wdict["chrom"],
                                        wdict["pos"],
                                        wdict["ref_base"],
                                        wdict["alt_base"],
                                    ) = self.liftover(
                                        wdict["chrom"],
                                        int(wdict["pos"]),
                                        wdict["ref_base"],
                                        wdict["alt_base"],
                                        lifter
                                    )
                                if not base_re.fullmatch(wdict["ref_base"]):
                                    raise IgnoredVariant("Invalid reference base")
                                if not base_re.fullmatch(wdict["alt_base"]):
                                    raise IgnoredVariant("Invalid alternate base")
                                p, r, a = (
                                    int(wdict["pos"]),
                                    wdict["ref_base"],
                                    wdict["alt_base"],
                                )
                                (
                                    new_pos,
                                    new_ref,
                                    new_alt,
                                ) = standardize_pos_ref_alt("+", p, r, a)
                                wdict["pos"] = new_pos
                                wdict["ref_base"] = new_ref
                                wdict["alt_base"] = new_alt
                                unique, UID = self.vtracker.addVar(
                                    wdict["chrom"], new_pos, new_ref, new_alt
                                )
                                wdict["uid"] = UID
                                if wdict["ref_base"] == wdict["alt_base"]:
                                    raise NoVariantError()
                                if unique:
                                    write_lnum += 1
                                    self.crv_writer.write_data(wdict)
                                    prelift_wdict["uid"] = UID
                                    self.crl_writer.write_data(prelift_wdict)
                                    # addl_operation errors shouldnt prevent variant from writing
                                    try:
                                        converter.addl_operation_for_unique_variant(  # type: ignore
                                            wdict, no_unique_var
                                        )
                                    except Exception as e:
                                        self._log_conversion_error(
                                            read_lnum, l, e, full_line_error=False
                                        )
                                    no_unique_var += 1
                                if UID not in UIDMap:
                                    # For this input line, only write to the .crm if the UID has not yet been written to the map file.
                                    self.crm_writer.write_data(
                                        {
                                            "original_line": read_lnum,
                                            "tags": wdict["tags"],
                                            "uid": UID,
                                            "fileno": f"{self.input_path_dict2[fname]}",
                                        }
                                    )
                                    UIDMap.append(UID)
                            self.crs_writer.write_data(wdict)
                    else:
                        raise IgnoredVariant("No valid alternate allele was found in any samples.")
                except Exception as e:
                    self._log_conversion_error(read_lnum, l, e)
                    continue
            f.close()
            cur_time = time()
            if total_lnum % 10000 == 0 or cur_time - last_status_update_time > 3:
                if self.status_writer is not None:
                    self.status_writer.queue_status_update(
                        "status",
                        "Running {} ({}): line {}".format(
                            "Converter", cur_fname, last_read_lnum
                        ),
                    )
                last_status_update_time = cur_time
            self.logger.info("error lines: %d" % self.file_error_lines)
        self._close_files()
        self.end()
        self.logger.info("total error lines: %d" % self.total_error_lines)
        if self.status_writer is not None:
            self.status_writer.queue_status_update("num_input_var", total_lnum)
            self.status_writer.queue_status_update("num_unique_var", write_lnum)
            self.status_writer.queue_status_update("num_error_input", self.total_error_lines)
        end_time = time()
        self.logger.info("finished: %s" % asctime(localtime(end_time)))
        runtime = round(end_time - start_time, 3)
        self.logger.info("num input lines: {}".format(total_lnum))
        self.logger.info("runtime: %s" % runtime)
        if self.status_writer is not None:
            self.status_writer.queue_status_update(
                "status",
                "Finished {} ({})".format(
                    "Converter", self.primary_converter.format_name
                ),
            )
        return total_lnum, self.primary_converter.format_name, genome_assemblies

    def liftover_one_pos(self, chrom, pos, lifter):
        from oakvar.exceptions import SetupError

        if lifter is None:
            raise SetupError("no lifter")
        res = lifter.convert_coordinate(chrom, pos - 1)
        if res is None or len(res) == 0:
            res_prev = lifter.convert_coordinate(chrom, pos - 2)
            res_next = lifter.convert_coordinate(chrom, pos)
            if res_prev is not None and res_next is not None:
                if len(res_prev) == 1 and len(res_next) == 1:
                    pos_prev = res_prev[0][1]
                    pos_next = res_next[0][1]
                    if pos_prev == pos_next - 2:
                        res = [(res_prev[0][0], pos_prev + 1)]
                    elif pos_prev == pos_next + 2:
                        res = [(res_prev[0][0], pos_prev - 1)]
        return res

    def liftover(self, chrom, pos, ref, alt, lifter):
        from oakvar.exceptions import LiftoverFailure
        from oakvar.util.util import reverse_complement
        from oakvar.exceptions import SetupError

        if lifter is None or self.wgsreader is None:
            raise SetupError("no lifter")
        reflen = len(ref)
        altlen = len(alt)
        if reflen == 1 and altlen == 1:
            res = self.liftover_one_pos(chrom, pos, lifter)
            if res is None or len(res) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res) > 1:
                raise LiftoverFailure("Liftover failure")
            try:
                el = res[0]
            except:
                raise LiftoverFailure("Liftover failure")
            newchrom = el[0]
            newpos = el[1] + 1
        elif reflen >= 1 and altlen == 0:  # del
            pos1 = pos
            pos2 = pos + reflen - 1
            res1 = lifter.convert_coordinate(chrom, pos1 - 1)
            res2 = lifter.convert_coordinate(chrom, pos2 - 1)
            if res1 is None or res2 is None or len(res1) == 0 or len(res2) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res1) > 1 or len(res2) > 1:
                raise LiftoverFailure("Liftover failure")
            el1 = res1[0]
            el2 = res2[0]
            newchrom1 = el1[0]
            newpos1 = el1[1] + 1
            newpos2 = el2[1] + 1
            newchrom = newchrom1
            newpos = newpos1
            newpos = min(newpos1, newpos2)
        elif reflen == 0 and altlen >= 1:  # ins
            res = lifter.convert_coordinate(chrom, pos - 1)
            if res is None or len(res) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res) > 1:
                raise LiftoverFailure("Liftover failure")
            el = res[0]
            newchrom = el[0]
            newpos = el[1] + 1
        else:
            pos1 = pos
            pos2 = pos + reflen - 1
            res1 = lifter.convert_coordinate(chrom, pos1 - 1)
            res2 = lifter.convert_coordinate(chrom, pos2 - 1)
            if res1 is None or res2 is None or len(res1) == 0 or len(res2) == 0:
                raise LiftoverFailure("Liftover failure")
            if len(res1) > 1 or len(res2) > 1:
                raise LiftoverFailure("Liftover failure")
            el1 = res1[0]
            el2 = res2[0]
            newchrom1 = el1[0]
            newpos1 = el1[1] + 1
            newpos2 = el2[1] + 1
            newchrom = newchrom1
            newpos = min(newpos1, newpos2)
        hg38_ref = self.wgsreader.get_bases(newchrom, newpos)
        if hg38_ref == reverse_complement(ref):
            newref = hg38_ref
            newalt = reverse_complement(alt)
        else:
            newref = ref
            newalt = alt
        return [newchrom, newpos, newref, newalt]

    def _log_conversion_error(self, ln, line, e, full_line_error=True):
        """Log exceptions thrown by primary converter.
        All exceptions are written to the .err file with the exception type
        and message. Exceptions are also written to the log file once, with the
        traceback.
        """
        from oakvar.exceptions import IgnoredVariant
        _ = line
        if not self.logger or not self.error_logger:
            from oakvar.exceptions import LoggerError

            raise LoggerError()
        from traceback import format_exc

        if full_line_error:
            self.file_error_lines += 1
            self.total_error_lines += 1
        err_str = format_exc().rstrip()
        if err_str not in self.unique_excs:
            self.unique_excs.append(err_str)
            if isinstance(e, IgnoredVariant) or (hasattr(e, "traceback") and e.traceback == False):
                pass
            else:
                self.logger.error(err_str)
        self.error_logger.error(f"{self.cur_file}:{ln}\t{str(e)}")
        # self.error_logger.error(
        #    "\nLINE:{:d}\nINPUT:{}\nERROR:{}\n#".format(ln, line[:-1], str(e))
        # )

    def _close_files(self):
        """Close the input and output files."""
        if self.crv_writer is not None:
            self.crv_writer.close()
        if self.crm_writer is not None:
            self.crm_writer.close()
        if self.crs_writer is not None:
            self.crs_writer.close()

    def end(self):
        pass


def main():
    master_cravat_converter = MasterCravatConverter()
    master_cravat_converter.run()


if __name__ == "__main__":
    master_cravat_converter = MasterCravatConverter()
    master_cravat_converter.run()



import markdown


def get_remote_manifest(handler):
    manifest = handler.get_remote_manifest()
    return manifest


def get_module_readme(request):
    from aiohttp.web import Response
    from ..module import get_readme

    module_name = request.match_info["module"]
    version = request.match_info["version"]
    if version == "latest":
        version = None
    readme_md = get_readme(module_name)
    if readme_md is None:
        response = Response(status=404)
    else:
        readme_html = markdown.markdown(readme_md)
        response = Response(body=readme_html, content_type="text/html")
    return response


def get_local_manifest():
    from aiohttp.web import json_response
    from ..module import list_local
    from ..module.local import get_local_module_info

    module_names = list_local()
    out = {}
    for module_name in module_names:
        local_info = get_local_module_info(module_name)
        if local_info is not None:
            out[module_name] = {
                "version": local_info.version,
                "type": local_info.type,
                "title": local_info.title,
                "description": local_info.description,
                "developer": local_info.developer,
            }
    return json_response(out)


def install_module(request):
    from aiohttp.web import Response
    from ..module import install_module

    module = request.json()
    module_name = module["name"]
    version = module["version"]
    install_module(module_name, version=version)  # , verbose=False)
    return Response()


def uninstall_module(request):
    from aiohttp.web import Response
    from ..module import uninstall_module

    module = request.json()
    module_name = module["name"]
    uninstall_module(module_name)
    return Response()



#!/usr/bin/env python3
import sys

if sys.platform == "win32" and sys.version_info >= (3, 8):
    from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy

    set_event_loop_policy(WindowsSelectorEventLoopPolicy())


class FilterColumn(object):

    test2sql = {
        "equals": "==",
        "lessThanEq": "<=",
        "lessThan": "<",
        "greaterThanEq": ">=",
        "greaterThan": ">",
        "hasData": "is not null",
        "noData": "is null",
        "stringContains": "like",
        "stringStarts": "like",  # Deprecated. Eliminate later
        "stringEnds": "like",  # Deprecated. Eliminate later
        "between": "between",
        "in": "in",
        "select": "in",
    }

    def __init__(self, d, parent_operator):
        self.column = d["column"]
        if self.column == "base__numsample":
            self.column = "tagsampler__numsample"
        self.test = d["test"]
        self.value = d.get("value")
        self.negate = d.get("negate", False)
        self.parent_operator = parent_operator

    def __repr__(self):
        return f"{self.column} {self.test} {self.value}"

    def get_sql(self):
        s = ""
        # TODO unify this to a single if/else on self.test
        if self.test == "multicategory":
            s = '{} like "%{}%"'.format(self.column, self.value[0])
            for v in self.value[1:]:
                s += ' or {} like "%{}%"'.format(self.column, v)
        elif self.test in ("select", "in"):
            ss = []
            for val in self.value:
                if type(val) == str:
                    val = '"{}"'.format(val)
                else:
                    val = str(val)
                ss.append(f"({self.column} = {val})")
            if ss:
                s = "(" + f" or ".join(ss) + ")"
            else:
                s = ""
        else:
            s = "{col} {opr}".format(col=self.column, opr=self.test2sql[self.test])
            sql_val = None
            if self.test == "equals":
                if type(self.value) is list:
                    v = self.value[0]
                    if type(v) is str:
                        sql_val = '"' + v + '"'
                    else:
                        sql_val = str(v)
                    for v in self.value[1:]:
                        if type(v) is str:
                            v = '"' + v + '"'
                        else:
                            v = str(v)
                        sql_val += " OR {} == {}".format(self.column, v)
                else:
                    if type(self.value) is str:
                        sql_val = '"{}"'.format(self.value)
                    else:
                        sql_val = str(self.value)
            elif self.test == "stringContains":
                sql_val = '"%{}%"'.format(self.value)
            elif self.test == "stringStarts":
                sql_val = '"{}%"'.format(self.value)
            elif self.test == "stringEnds":
                sql_val = '"%{}"'.format(self.value)
            elif self.test == "between":
                sql_val = "{} and {}".format(self.value[0], self.value[1])
            elif self.test in (
                "lessThan",
                "lessThanEq",
                "greaterThan",
                "greaterThanEq",
            ):
                sql_val = str(self.value)
            if sql_val:
                s += " (" + sql_val + ")"
        if self.negate:
            s = "not(" + s + ")"
        return s


class FilterGroup(object):
    def __init__(self, d):
        self.operator = d.get("operator", "and")
        self.negate = d.get("negate", False)
        self.rules = []
        for rule in d.get("rules", []):
            if "operator" in rule:
                self.rules.append(FilterGroup(rule))
            else:
                self.rules.append(FilterColumn(rule, self.operator))
        # Backwards compatability, may remove later
        self.rules += [FilterGroup(x) for x in d.get("groups", [])]
        self.rules += [FilterColumn(x, self.operator) for x in d.get("columns", [])]
        self.column_prefixes = {}

    def add_prefixes(self, prefixes):
        self.column_prefixes.update(prefixes)
        for rule in self.rules:
            if isinstance(rule, FilterGroup):
                rule.add_prefixes(prefixes)
            elif isinstance(rule, FilterColumn):
                prefix = self.column_prefixes.get(rule.column)
                if prefix is not None:
                    rule.column = prefix + "." + rule.column

    def get_sql(self):
        clauses = []
        for operand in self.rules:
            clause = operand.get_sql()
            if clause:
                clauses.append(clause)
        s = ""
        if clauses:
            s += "("
            sql_operator = " " + self.operator + " "
            s += sql_operator.join(clauses)
            s += ")"
            if self.negate:
                s = "not" + s
        return s


class CravatFilter:
    @classmethod
    async def create(
        cls,
        dbpath=None,
        filterpath=None,
        filtername=None,
        filterstring=None,
        filter=None,
        mode="sub",
        filtersql=None,
        includesample=None,
        excludesample=None,
    ):
        self = CravatFilter(
            dbpath=dbpath,
            filterpath=filterpath,
            filtername=filtername,
            filterstring=filterstring,
            filter=filter,
            mode=mode,
            filtersql=filtersql,
            includesample=includesample,
            excludesample=excludesample,
        )
        await self.second_init()
        return self

    def __init__(
        self,
        dbpath=None,
        filterpath=None,
        filtername=None,
        filterstring=None,
        filter=None,
        filtersql=None,
        includesample=None,
        excludesample=None,
        mode="sub",
    ):
        self.mode = mode
        if self.mode == "main":
            self.stdout = True
        else:
            self.stdout = False
        self.dbpath = dbpath
        self.conn = None
        self.filterpath = filterpath
        self.cmd = None
        self.level = None
        self.filter = filter
        self.savefiltername = None
        self.filtername = None
        self.filterstring = filterstring
        self.filtersql = filtersql
        self.includesample = includesample
        self.excludesample = excludesample
        if filter != None:
            self.filter = filter
        else:
            if filterstring != None:
                self.filterstring = filterstring
            elif filtername != None:
                self.filtername = filtername
            elif filterpath != None:
                self.filterpath = filterpath
        self.filtertable = "filter"
        self.generows = {}
        self.table_aliases = None
        self.column_prefixes = {}

    async def get_module_version_in_job(self, module_name, conn=None, cursor=None):
        if conn is None or cursor is None:
            return None
        q = 'select colval from info where colkey="_annotators"'
        await cursor.execute(q)
        anno_vers = await cursor.fetchone()
        if anno_vers is None:
            return None
        version = None
        for anno_ver in anno_vers[0].split(","):
            [mname, ver] = anno_ver.split(":")
            if mname == module_name:
                version = ver
                break
        return version

    async def exec_db(self, func, *args, **kwargs):
        conn = await self.get_db_conn()
        ret = None
        if conn is not None:
            cursor = await conn.cursor()
            ret = await func(*args, conn=conn, cursor=cursor, **kwargs)
            await cursor.close()
        return ret

    async def second_init(self):
        if self.mode == "sub":
            if self.dbpath != None:
                await self.connect_db()
            await self.exec_db(self.loadfilter)

    async def run(self, cmd=None, args=None, dbpath=None, filter=None):
        if args != None:
            self.parse_args(args)
        if cmd != None:
            self.cmd = cmd

        if dbpath != None:
            self.dbpath = dbpath

        # Loads filter.
        if filter != None:
            self.filter = filter
        elif (
            self.filtername != None
            or self.filterpath != None
            or self.filterstring != None
            or self.filtersql != None
        ) and self.filter == None:
            await self.exec_db(self.loadfilter)

        ret = None
        if self.dbpath is not None:
            if self.cmd == "count":
                ret = await self.run_level_based_func(self.exec_db(self.getcount))
            elif self.cmd == "rows":
                ret = await self.run_level_based_func(self.exec_db(self.getrows))
            elif self.cmd == "pipe":
                ret = await self.run_level_based_func(self.exec_db(self.getiterator))
        elif self.dbpath is not None and self.cmd == "list":
            ret = self.exec_db(self.listfilter)

        # Saves filter.
        if self.filter != None:
            if self.cmd == "save" or self.savefiltername != None:
                ret = self.exec_db(self.savefilter)

        return ret

    async def run_level_based_func(self, cmd):
        ret = {}
        if self.level != None:
            ret[self.level] = await cmd(level=self.level)
        else:
            levels = ["variant", "gene"]
            ret = {}
            for level in levels:
                ret_onelevel = await cmd(level=level)
                ret[level] = ret_onelevel
        return ret

    def parse_args(self, args):
        from argparse import ArgumentParser

        parser = ArgumentParser()
        parser.add_argument(
            "-d",
            dest="dbpath",
            required=True,
            help="Path of a result database file (.sqlite)",
        )
        parser.add_argument(
            "-f", dest="filterpath", help="Path of a filtering criteria file"
        )
        parser.add_argument(
            "-F",
            dest="filtername",
            help="Name of the filter to apply (saved in the database)",
        )
        parser.add_argument(
            "--filterstring", dest="filterstring", default=None, help="Filter in JSON"
        )
        parser.add_argument(
            "-l",
            dest="level",
            default=None,
            choices=["variant", "gene"],
            help="Analysis level to filter",
        )
        parser.add_argument(
            "-s",
            dest="savefiltername",
            help="Name to save the filter as (in the database)",
        )
        parser.add_argument(
            "--filtersql", dest="filtersql", default=None, help="Filter SQL"
        )
        parser.add_argument(
            "--includesample",
            dest="includesample",
            nargs="+",
            default=None,
            help="Sample IDs to include",
        )
        parser.add_argument(
            "--excludesample",
            dest="excludesample",
            nargs="+",
            default=None,
            help="Sample IDs to exclude",
        )
        if self.mode == "main":
            parser.add_argument(
                "command",
                choices=["uidpipe", "count", "rows", "pipe", "save", "list"],
                help="Command",
            )

        parsed_args = parser.parse_args(args)
        self.dbpath = parsed_args.dbpath
        self.filterpath = parsed_args.filterpath
        self.level = parsed_args.level
        if self.mode == "main":
            self.cmd = parsed_args.command
        self.savefiltername = parsed_args.savefiltername
        self.filtername = parsed_args.filtername
        self.filterstring = parsed_args.filterstring
        self.filtersql = parsed_args.filtersql

    async def get_db_conn(self):
        if self.dbpath is None:
            return None
        if self.conn is None:
            import aiosqlite

            self.conn = await aiosqlite.connect(self.dbpath)
        return self.conn

    async def connect_db(self, dbpath=None):
        if dbpath != None:
            self.dbpath = dbpath
        conn = await self.get_db_conn()
        if conn is not None:
            await conn.create_function("regexp", 2, regexp)
            await self.exec_db(self.set_aliases)

    async def set_aliases(self, conn=None, cursor=None):
        if conn is None:
            pass
        self.table_aliases = {"variant": "v", "gene": "g"}
        self.column_prefixes = {}
        q = "pragma table_info(variant)"
        if cursor is not None:
            await cursor.execute(q)
            self.column_prefixes.update(
                {
                    row[1]: self.table_aliases["variant"]
                    for row in await cursor.fetchall()
                }
            )
            q = "pragma table_info(gene)"
            await cursor.execute(q)
            self.column_prefixes.update(
                {row[1]: self.table_aliases["gene"] for row in await cursor.fetchall()}
            )

    async def close_db(self):
        if self.conn is not None:
            await self.conn.close()
            self.conn = None

    async def create_filtertable(self, conn=None, cursor=None):
        if conn is not None and cursor is not None:
            sql = "create table " + self.filtertable + " (name text, criteria text)"
            await cursor.execute(sql)
            await conn.commit()

    async def filtertable_exists(self, conn=None, cursor=None):
        ret = False
        if conn is not None and cursor is not None:
            sql = f'select * from viewersetup where datatype="filter" and name="{self.filtername}"'
            await cursor.execute(sql)
            row = await cursor.fetchone()
            if row == None:
                ret = False
            else:
                ret = True
        return ret

    async def loadfilter(
        self,
        filterpath=None,
        filtername=None,
        filterstring=None,
        filtersql=None,
        filter=None,
        includesample=None,
        excludesample=None,
        conn=None,
        cursor=None,
    ):
        if conn is None:
            pass
        from os.path import exists
        from oyaml import safe_load
        from json import load, loads

        if filterpath != None:
            self.filterpath = filterpath
        if filtername != None:
            self.filtername = filtername
        if filterstring != None:
            self.filterstring = filterstring
        if filtersql != None:
            self.filtersql = filtersql
        if filter != None:
            self.filter = filter
        if includesample is not None:
            self.includesample = includesample
        if excludesample is not None:
            self.excludesample = excludesample
        filter_table_present = await self.exec_db(self.filtertable_exists)
        if self.filter:
            pass
        elif self.filtersql is not None:
            if exists(self.filtersql):  # a file can be used.
                with open(self.filtersql) as f:
                    self.filtersql = "".join(f.readlines())
            self.filter = {}
        elif self.filterstring is not None:
            self.filterstring = self.filterstring.replace("'", '"')
            self.filter = loads(self.filterstring)
        elif (
            self.filtername is not None and filter_table_present and cursor is not None
        ):
            await cursor.execute(
                "select viewersetup from viewersetup"
                + ' where datatype="filter" and name="'
                + self.filtername
                + '"'
            )
            criteria = await cursor.fetchone()
            if criteria != None:
                self.filter = loads(criteria[0])["filterSet"]
        elif self.filterpath is not None and exists(self.filterpath):
            with open(self.filterpath) as f:
                ftype = self.filterpath.split(".")[-1]
                if ftype in ["yml", "yaml"]:
                    self.filter = safe_load(f)
                elif ftype in ["json"]:
                    self.filter = load(f)
        if self.filter is None:
            self.filter = {}
        await self.verify_filter(cursor)

    async def verify_filter(self, cursor):
        from oakvar.exceptions import InvalidFilter

        wrong_samples = await self.verify_filter_sample(cursor)
        wrong_colnames = await self.verify_filter_module(cursor)
        if len(wrong_samples) > 0 or len(wrong_colnames) > 0:
            raise InvalidFilter(wrong_samples, wrong_colnames)

    async def check_sample_name(self, sample_id, cursor):
        await cursor.execute(
            'select base__sample_id from sample where base__sample_id="'
            + sample_id
            + '" limit 1'
        )
        ret = await cursor.fetchone()
        return ret is not None

    async def verify_filter_sample(self, cursor):
        if "sample" not in self.filter:
            return []
        ft = self.filter["sample"]
        wrong_samples = set()
        if "require" in ft:
            for rq in ft["require"]:
                if await self.check_sample_name(rq, cursor) == False:
                    wrong_samples.add(rq)
        if "reject" in ft:
            for rq in ft["reject"]:
                if await self.check_sample_name(rq, cursor) == False:
                    wrong_samples.add(rq)
        return wrong_samples

    async def col_name_exists(self, colname, cursor):
        await cursor.execute(
            'select col_def from variant_header where col_name="'
            + colname
            + '" limit 1'
        )
        ret = await cursor.fetchone()
        if ret is None:
            await cursor.execute(
                'select col_def from gene_header where col_name="'
                + colname
                + '" limit 1'
            )
            ret = await cursor.fetchone()
        return ret is not None

    async def extract_filter_columns(self, rules, colset, cursor):
        for rule in rules:
            if "column" in rule:
                if await self.col_name_exists(rule["column"], cursor) == False:
                    colset.add(rule["column"])
            elif "rules" in rule:
                await self.extract_filter_columns(rule["rules"], colset, cursor)

    async def verify_filter_module(self, cursor):
        if "variant" not in self.filter:
            return []
        wrong_modules = set()
        if "rules" in self.filter["variant"]:
            await self.extract_filter_columns(
                self.filter["variant"]["rules"], wrong_modules, cursor
            )
        return wrong_modules

    async def delete_filtered_uid_table(self, conn=None, cursor=None):
        if conn is not None and cursor is not None:
            await cursor.execute("pragma synchronous=0")
            q = "drop table if exists variant_filtered"
            await cursor.execute(q)
            q = "drop table if exists gene_filtered"
            await cursor.execute(q)
            await conn.commit()
            await cursor.execute("pragma synchronous=2")

    def getwhere(self, level):
        where = ""
        if self.filter is not None and level in self.filter:
            criteria = self.filter[level]
            main_group = FilterGroup(criteria)
            main_group.add_prefixes(self.column_prefixes)
            sql = main_group.get_sql()
            if sql:
                where = "where " + sql
        return where

    def getvariantcount(self):
        from asyncio import get_event_loop

        loop = get_event_loop()
        count = loop.run_until_complete(self.exec_db(self.getcount, "variant"))
        return count

    def getgenecount(self):
        from asyncio import get_event_loop

        loop = get_event_loop()
        count = loop.run_until_complete(self.exec_db(self.getcount, "gene"))
        return count

    async def getcount(self, level="variant", conn=None, cursor=None):
        if conn is not None and cursor is not None:
            bypassfilter = (
                not self.filter
                and not self.filtersql
                and not self.includesample
                and not self.excludesample
            )
            level = "variant"
            await self.exec_db(self.make_filtered_uid_table)
            if bypassfilter:
                ftable = level
            else:
                ftable = level + "_filtered"
            q = "select count(*) from " + ftable
            await cursor.execute(q)
            n = None
            for row in await cursor.fetchone():
                n = row
            if self.stdout == True:
                print("#" + level)
                print(str(n))
            return n
        else:
            return None

    def getvariantrows(self):
        from asyncio import get_event_loop

        loop = get_event_loop()
        rows = loop.run_until_complete(self.exec_db(self.getrows, "variant"))
        return rows

    def getgenerows(self):
        from asyncio import get_event_loop

        loop = get_event_loop()
        rows = loop.run_until_complete(self.exec_db(self.getrows, "gene"))
        return rows

    async def getrows(self, level="variant", conn=None, cursor=None):
        ret = None
        if conn is not None and cursor is not None:
            if level != "variant":
                return
            where = self.getwhere(level)
            await self.exec_db(self.make_filtered_sample_table)
            await self.exec_db(self.make_gene_list_table)
            await conn.commit()
            q = f"select v.* from variant as v join fsample as s on v.base__uid=s.base__uid"
            if isinstance(self.filter, dict) and len(self.filter.get("genes", [])) > 0:
                q += " join gene_list as gl on v.base__hugo=gl.base__hugo"
            if "g." in where:
                q += " join gene as g on v.base__hugo=g.base__hugo"
            await cursor.execute(q)
            ret = [list(v) for v in await cursor.fetchall()]
            if self.stdout == True:
                print("#" + level)
                for row in ret:
                    print("\t".join([str(v) for v in row]))
        return ret

    async def make_generows(self, conn=None, cursor=None):
        if conn is None:
            pass
        if cursor is not None:
            q = "select * from gene"
            await cursor.execute(q)
            rows = await cursor.fetchall()
            self.generows = {}
            for row in rows:
                hugo = row[0]
                self.generows[hugo] = row

    async def get_gene_row(self, hugo):
        if hugo is None:
            return None
        if bool(self.generows) == False:
            await self.exec_db(self.make_generows)
        row = self.generows.get(hugo)
        return row

    def getvariantiterator(self):
        from asyncio import get_event_loop

        loop = get_event_loop()
        iterator = loop.run_until_complete(self.exec_db(self.getiterator, "variant"))
        return iterator

    def getgeneiterator(self):
        from asyncio import get_event_loop

        loop = get_event_loop()
        iterator = loop.run_until_complete(self.exec_db(self.getiterator, "gene"))
        return iterator

    async def getiterator(self, level="variant", conn=None, cursor=None):
        if conn is None:
            pass
        if cursor is not None:
            (
                __sample_needed__,
                __tag_needed__,
                include_where,
                exclude_where,
            ) = self.getwhere(level)
            sql = (
                "select *  from "
                + level
                + include_where
                + " except select * from "
                + level
                + exclude_where
            )
            await cursor.execute(sql)
            it = await cursor.fetchall()
            return it
        else:
            return None

    async def get_filtered_iterator(self, level="variant", page=None, pagesize=None, conn=None, cursor=None):
        bypassfilter = (
            not self.filter
            and not self.filtersql
            and not self.includesample
            and not self.excludesample
        )
        if conn is None:
            pass
        ftable = None
        kcol = None
        sql = None
        if level == "variant":
            kcol = "base__uid"
            if bypassfilter:
                ftable = "variant"
            else:
                ftable = "variant_filtered"
        elif level == "gene":
            kcol = "base__hugo"
            if bypassfilter:
                ftable = "gene"
            else:
                ftable = "gene_filtered"
        elif level == "sample":
            kcol = "base__uid"
            if bypassfilter:
                ftable = "variant"
            else:
                ftable = "variant_filtered"
        elif level == "mapping":
            kcol = "base__uid"
            if bypassfilter:
                ftable = "variant"
            else:
                ftable = "variant_filtered"
        table = level
        if level in ["variant", "gene", "sample", "mapping"]:
            if level == "gene" and bypassfilter:
                if cursor is not None:
                    sql = "pragma table_info(gene)"
                    await cursor.execute(sql)
                    rs = await cursor.fetchall()
                    colnames = ["gene." + r[1] for r in rs if r[1] != "base__hugo"]
                    sql = "select distinct variant.base__hugo, {} from variant inner join gene on variant.base__hugo==gene.base__hugo".format(
                        ", ".join(colnames)
                    )
            else:
                sql = "select v.* from " + table + " as v"
                if bypassfilter == False and ftable is not None and kcol is not None:
                    sql += " inner join " + ftable + " as f on v." + kcol + "=f." + kcol
        if not cursor:
            return None, None, None
        if not sql:
            return None, None, None
        if level == "variant" and pagesize is not None:
            if not page:
                page = 1
            offset = (page - 1) * pagesize
            final_sql = sql + f" limit {pagesize} offset {offset}"
        else:
            final_sql = sql
        await cursor.execute(final_sql)
        cols = [v[0] for v in cursor.description]
        rows = await cursor.fetchall()
        if level == "variant":
            norows_q = sql.replace("v.*", "count(*)")
            await cursor.execute(norows_q)
            ret = await cursor.fetchone()
            total_norows = ret[0]
        else:
            total_norows = len(rows)
        return cols, rows, total_norows

    async def make_filtered_sample_table(self, conn=None, cursor=None):
        if conn is not None and cursor is not None:
            q = "drop table if exists fsample"
            await cursor.execute(q)
            await conn.commit()
            req = []
            rej = []
            if "sample" in self.filter:
                if "require" in self.filter["sample"]:
                    req = self.filter["sample"]["require"]
                if "reject" in self.filter["sample"]:
                    rej = self.filter["sample"]["reject"]
            if self.includesample is not None:
                req = self.includesample
            if self.excludesample is not None:
                rej = self.excludesample
            if len(req) > 0 or len(rej) > 0:
                q = "create table fsample as select distinct base__uid from sample"
                if req:
                    q += " where base__sample_id in ({})".format(
                        ", ".join(['"{}"'.format(sid) for sid in req])
                    )
                for s in rej:
                    q += ' except select base__uid from sample where base__sample_id="{}"'.format(
                        s
                    )
                await cursor.execute(q)
                await conn.commit()
                return True
            else:
                return False
        else:
            return False

    async def make_filter_where(self, conn=None, cursor=None):
        if conn is None or cursor is None:
            pass
        q = ""
        if len(self.filter) == 0:
            if self.includesample is not None or self.excludesample is not None:
                fsample_made = await self.exec_db(self.make_filtered_sample_table)
            else:
                fsample_made = False
            if self.filtersql is not None:
                if "g." in self.filtersql:
                    q += " join gene as g on v.base__hugo=g.base__hugo"
                if "s." in self.filtersql:
                    q += " join sample as s on v.base__uid=s.base__uid"
                q += " where " + self.filtersql
            if fsample_made:
                q += " and (v.base__uid in (select base__uid from fsample))"
        else:
            where = self.getwhere("variant")
            fsample_made = await self.exec_db(self.make_filtered_sample_table)
            gene_list_made = await self.exec_db(self.make_gene_list_table)
            if gene_list_made:
                if (
                    isinstance(self.filter, dict)
                    and len(self.filter.get("genes", [])) > 0
                ):
                    q += " join gene_list as gl on v.base__hugo=gl.base__hugo"
            if "g." in where:
                q += " join gene as g on v.base__hugo=g.base__hugo"
            q += " " + where
            if fsample_made:
                q += " and (v.base__uid in (select base__uid from fsample))"
        return q

    async def make_filtered_uid_table(self, conn=None, cursor=None):
        bypassfilter = (
            not self.filter
            and not self.filtersql
            and not self.includesample
            and not self.excludesample
        )
        if not conn or not cursor:
            from ..exceptions import DatabaseConnectionError

            raise DatabaseConnectionError()
        level = "variant"
        vftable = level + "_filtered"
        q = "drop table if exists " + vftable
        await cursor.execute(q)
        q = f"create table {vftable} as select v.base__uid from variant as v"
        if not bypassfilter:
            if not self.filter:
                if self.includesample or self.excludesample:
                    fsample_made = await self.exec_db(self.make_filtered_sample_table)
                else:
                    fsample_made = False
                if self.filtersql is not None:
                    if "g." in self.filtersql:
                        q += " join gene as g on v.base__hugo=g.base__hugo"
                    if "s." in self.filtersql:
                        q += " join sample as s on v.base__uid=s.base__uid"
                    q += " where " + self.filtersql
                if fsample_made:
                    if " where " in q:
                        q += " and "
                    else:
                        q += " where "
                    q += "(v.base__uid in (select base__uid from fsample))"
            else:
                where = self.getwhere(level)
                fsample_made = await self.exec_db(self.make_filtered_sample_table)
                gene_list_made = await self.exec_db(self.make_gene_list_table)
                if gene_list_made:
                    if (
                        isinstance(self.filter, dict)
                        and len(self.filter.get("genes", [])) > 0
                    ):
                        q += " join gene_list as gl on v.base__hugo=gl.base__hugo"
                if "g." in where:
                    q += " join gene as g on v.base__hugo=g.base__hugo"
                q += " " + where
                if fsample_made:
                    if " where " in q:
                        q += " and "
                    else:
                        q += " where "
                    q += "(v.base__uid in (select base__uid from fsample))"
        await cursor.execute(q)
        await conn.commit()

    async def make_gene_list_table(self, conn=None, cursor=None):
        if conn is not None and cursor is not None:
            tname = "gene_list"
            q = "drop table if exists {}".format(tname)
            await cursor.execute(q)
            await conn.commit()
            if isinstance(self.filter, dict) and "genes" in self.filter:
                tdata = [(hugo,) for hugo in self.filter["genes"]]
            else:
                tdata = []
            if len(tdata) > 0:
                q = "create table {} (base__hugo text)".format(tname)
                await cursor.execute(q)
                q = "insert into {} (base__hugo) values (?)".format(tname)
                await cursor.executemany(q, tdata)
                await conn.commit()
                return True
            else:
                return False
        else:
            return False

    async def make_filtered_hugo_table(self, conn=None, cursor=None):
        bypassfilter = (
            not self.filter
            and not self.filtersql
            and not self.includesample
            and not self.excludesample
        )
        if conn is not None and cursor is not None and bypassfilter == False:
            await cursor.execute("pragma synchronous=0")
            vftable = "variant_filtered"
            gftable = "gene_filtered"
            q = "drop table if exists " + gftable
            await cursor.execute(q)
            q = (
                "create table "
                + gftable
                + " as select distinct v.base__hugo from variant as v"
                " inner join " + vftable + " as vf on vf.base__uid=v.base__uid"
                " where v.base__hugo is not null"
            )
            await cursor.execute(q)
            await cursor.execute("pragma synchronous=2")

    async def savefilter(self, name=None, conn=None, cursor=None):
        if conn is not None and cursor is not None:
            from json import dumps

            if conn is None or self.filter is None:
                return
            if name == None:
                if self.savefiltername != None:
                    name = self.savefiltername
                else:
                    name = "default"
            # Creates filter save table if not exists.
            await cursor.execute(
                "select name from sqlite_master where "
                + 'type="table" and name="'
                + self.filtertable
                + '"'
            )
            for ret in await cursor.fetchone():
                if ret == None:
                    await cursor.execute(
                        "create table "
                        + self.filtertable
                        + " (name text unique, criteria text)"
                    )
            # Saves the filter.
            filterstr = dumps(self.filter)
            sql = (
                "insert or replace into "
                + self.filtertable
                + ' values ("'
                + name
                + "\", '"
                + filterstr
                + "')"
            )
            await cursor.execute(sql)
            await conn.commit()

    async def listfilter(self, name=None, conn=None, cursor=None):
        if conn is None:
            pass
        ret = {}
        if cursor is not None:
            from json import loads

            if name == None:
                if self.savefiltername != None:
                    name = self.savefiltername
                else:
                    name = "default"
            # Creates filter save table if not exists.
            await cursor.execute(
                "select name from sqlite_master where "
                + 'type="table" and name="'
                + self.filtertable
                + '"'
            )
            for ret in await cursor.fetchone():
                if ret == None:
                    await cursor.execute(
                        "create table "
                        + self.filtertable
                        + " (name text, criteria text)"
                    )
            sql = "select name, criteria from " + self.filtertable
            await cursor.execute(sql)
            for row in await cursor.fetchall():
                name = row[0]
                criteria = loads(row[1])
                ret[name] = criteria
                if self.stdout:
                    print("#" + name)
                    for level in criteria:
                        print("    " + level + ":")
                        for column in criteria[level]:
                            print("        " + column + ": " + criteria[level][column])
        return ret

    def addvariantfilter(self, column, condition):
        self.addfilter(column, condition, "variant")

    def addgenefilter(self, column, condition):
        self.addfilter(column, condition, "gene")

    def addfilter(self, column, condition, level="variant"):
        if self.filter == None:
            self.filter = {}
        if level not in self.filter:
            self.filter[level] = {}
        self.filter[level][column] = condition

    def removevariantfilter(self, column):
        self.removefilter(column, "variant")

    def removegenefilter(self, column):
        self.removefilter(column, "gene")

    def removefilter(self, column, level="variant"):
        if self.filter == None:
            return
        if level in self.filter and column in self.filter[level]:
            del self.filter[level][column]

    async def table_exists(self, table, conn=None, cursor=None):
        if conn is None:
            pass
        if cursor is not None:
            sql = (
                'select name from sqlite_master where type="table" and '
                + 'name="'
                + table
                + '"'
            )
            await cursor.execute(sql)
            for row in await cursor.fetchone():
                if row == None:
                    return False
                else:
                    return True
        else:
            return False

    async def get_variant_iterator_filtered_uids_cols(
        self, cols, conn=None, cursor=None
    ):
        if conn is None:
            pass
        if cursor is not None:
            if cols[0] == "base__uid":
                cols[0] = "v." + cols[0]
            bypassfilter = (
                not self.filter
                and not self.filtersql
                and not self.includesample
                and not self.excludesample
            )
            q = "select " + ",".join(cols) + " from variant as v"
            if bypassfilter == False:
                q += " inner join variant_filtered as f on v.base__uid=f.base__uid"
            await cursor.execute(q)
            rows = await cursor.fetchall()
            for row in rows:
                d = {}
                for i in range(len(row)):
                    d[cols[i].split("__")[1]] = row[i]
                yield d

    async def get_filtered_hugo_list(self, conn=None, cursor=None):
        if conn is None:
            pass
        if cursor:
            bypassfilter = (
                not self.filter
                and not self.filtersql
                and not self.includesample
                and not self.excludesample
            )
            if bypassfilter:
                q = "select distinct variant.base__hugo from gene, variant where gene.base__hugo==variant.base__hugo"
            else:
                q = "select base__hugo from gene_filtered"
            await cursor.execute(q)
            rows = await cursor.fetchall()
            hugos = [row[0] for row in rows]
            return hugos

    async def get_variant_data_for_cols(self, cols, conn=None, cursor=None):
        if conn is None:
            pass
        if cursor is not None:
            bypassfilter = (
                not self.filter
                and not self.filtersql
                and not self.includesample
                and not self.excludesample
            )
            if cols[0] == "base__uid":
                cols[0] = "v.base__uid"
            q = "select {},base__hugo from variant as v".format(",".join(cols))
            if bypassfilter == False:
                q += " inner join variant_filtered as f on v.base__uid=f.base__uid"
            if cols[0] == "v.base__uid":
                cols[0] = "base__uid"
            await cursor.execute(q)
            rows = await cursor.fetchall()
            return rows

    async def get_variant_data_for_hugo(self, hugo, cols, conn=None, cursor=None):
        if conn is None:
            pass
        if cursor is not None:
            bypassfilter = (
                not self.filter
                and not self.filtersql
                and not self.includesample
                and not self.excludesample
            )
            if cols[0] == "base__uid":
                cols[0] = "v.base__uid"
            q = "select {} from variant as v".format(",".join(cols))
            if bypassfilter == False:
                q += ' inner join variant_filtered as f on v.base__uid=f.base__uid and v.base__hugo="{}"'.format(
                    hugo
                )
            if cols[0] == "v.base__uid":
                cols[0] = "base__uid"
            await cursor.execute(q)
            rows = await cursor.fetchall()
            return rows

    async def get_result_levels(self, conn=None, cursor=None):
        if conn is None:
            pass
        table_names = []
        if cursor is not None:
            q = (
                'select name from sqlite_master where type="table" and '
                + 'name like "%_header"'
            )
            await cursor.execute(q)
            for row in await cursor.fetchall():
                table_names.append(row[0].replace("_header", ""))
        return table_names

    async def get_stored_output_columns(self, module_name, conn=None, cursor=None):
        if conn is None:
            pass
        output_columns = []
        if cursor is not None:
            from json import loads

            q = f'select col_def from variant_header where col_name like "{module_name}\\_\\_%" escape "\\"'
            await cursor.execute(q)
            for row in await cursor.fetchall():
                d = loads(row[0])
                d["name"] = d["name"].replace(f"{module_name}__", "")
                output_columns.append(d)
        return output_columns


def regexp(y, x, search=None):
    import re

    if search is None:
        search = re.search
    if x is None:
        return 0
    return 1 if search(y, x) else 0


def main():
    import sys
    from asyncio import new_event_loop

    loop = new_event_loop()
    cv = loop.run_until_complete(CravatFilter.create(mode="main"))
    loop.run_until_complete(cv.run(args=sys.argv[1:]))
    loop.close()



from ..decorators import cli_func
from ..decorators import cli_entry


@cli_entry
def cli_run(args):
    ret = run(args)
    return ret


@cli_func
def run(args, __name__="run"):
    from asyncio import run
    from sys import platform
    from sys import version_info
    import nest_asyncio
    from ..system import custom_system_conf

    # nested asyncio
    nest_asyncio.apply()
    # Windows event loop patch
    if platform == "win32" and version_info >= (3, 8):
        from asyncio import set_event_loop_policy, WindowsSelectorEventLoopPolicy  # type: ignore

        set_event_loop_policy(WindowsSelectorEventLoopPolicy())
    # Custom system conf
    custom_system_conf = {}
    system_option = args.get("system_option")
    if system_option:
        for kv in system_option:
            if "=" not in kv:
                continue
            toks = kv.split("=")
            if len(toks) != 2:
                continue
            [k, v] = toks
            try:
                v = int(v)
            except ValueError:
                pass
            custom_system_conf[k] = v
    module = Cravat(**args)
    return run(module.main())


class Cravat(object):
    def __init__(self, **kwargs):
        from sys import executable

        self.runlevels = {
            "converter": 1,
            "mapper": 2,
            "annotator": 3,
            "aggregator": 4,
            "postaggregator": 5,
            "reporter": 6,
        }
        self.should_run_converter = False
        self.should_run_genemapper = False
        self.should_run_annotators = True
        self.should_run_aggregator = True
        self.should_run_reporter = True
        self.mapper_ran = False
        self.annotator_ran = False
        self.aggregator_ran = False
        self.done_annotators = {}
        self.status_json_path = None
        self.status_json = None
        self.pkg_ver = None
        self.logger = None
        self.logmode = "w"
        self.log_path = None
        self.error_logger = None
        self.log_handler = None
        self.error_log_handler = None
        self.start_time = None
        self.unique_logs = None
        self.manager = None
        self.status_writer = None
        self.run_annotators = {}
        self.result_path = None
        self.package_conf = {}
        self.args = None
        self.main_conf = {}
        self.conf_run = {}
        self.conf_path = None
        self.conf = {}
        self.num_input = None
        self.first_non_url_input = None
        self.inputs = None
        self.run_name = None
        self.output_dir = None
        self.startlevel = self.runlevels["converter"]
        self.endlevel = self.runlevels["postaggregator"]
        self.verbose = False
        self.cleandb = False
        self.excludes = []
        self.annotator_names = []
        self.mapper_name = None
        self.mapper = None
        self.postaggregator_names = []
        self.postaggregators = {}
        self.report_names = []
        self.reports = {}
        self.reporter_names = []
        self.crvinput = None
        self.crxinput = None
        self.crginput = None
        self.crv_present = False
        self.crx_present = False
        self.crg_present = False
        self.annot_names = []
        self.numinput = None
        self.converter_format = None
        self.genemapper = None
        self.ordered_summarizers = []
        self.pythonpath = executable
        self.annotators = {}
        self.append_mode = False
        self.pipeinput = False
        self.exception = None
        self.genome_assembiles = None
        self.inkwargs = kwargs

    def check_valid_modules(self, module_names):
        from ..exceptions import ModuleNotExist
        from ..module.local import module_exists_local

        for module_name in module_names:
            if not module_exists_local(module_name):
                raise ModuleNotExist(module_name)

    def close_logger(self):
        import logging

        if self.log_handler:
            self.log_handler.close()
            if self.logger is not None:
                self.logger.removeHandler(self.log_handler)
        if self.error_log_handler:
            self.error_log_handler.close()
            if self.error_logger:
                self.error_logger.removeHandler(self.error_log_handler)
        logging.shutdown()

    def delete_output_files(self):
        from ..exceptions import SetupError
        from ..util.util import quiet_print
        import os
        import glob

        if self.run_name is None or self.output_dir is None:
            raise SetupError()
        fns = glob.glob(os.path.join(self.output_dir, self.run_name + ".*"))
        for fn in fns:
            quiet_print(f"  Removing {fn}", self.args)
            os.remove(fn)

    def download_url_input(self, input_no):
        from ..exceptions import NoInput
        from ..util.util import is_url, humanize_bytes
        from ..util.util import quiet_print
        import os
        import requests

        if self.inputs is None:
            raise NoInput()
        ip = self.inputs[input_no]
        if " " in ip:
            print(f"Space is not allowed in input file paths ({ip})")
            exit()
        if is_url(ip):
            quiet_print(f"Fetching {ip}... ", self.args)
            try:
                r = requests.head(ip)
                r = requests.get(ip, stream=True)
                fn = os.path.basename(ip)
                fpath = fn
                cur_size = 0.0
                num_total_star = 40.0
                total_size = float(r.headers["content-length"])
                with open(fpath, "wb") as wf:
                    for chunk in r.iter_content(chunk_size=8192):
                        wf.write(chunk)
                        cur_size += float(len(chunk))
                        perc = cur_size / total_size
                        cur_star = int(perc * num_total_star)
                        rem_stars = int(num_total_star - cur_star)
                        cur_prog = "*" * cur_star
                        rem_prog = " " * rem_stars
                        print(
                            f"[{cur_prog}{rem_prog}] {humanize_bytes(cur_size)} / {humanize_bytes(total_size)} ({perc * 100.0:.0f}%)",
                            end="\r",
                            flush=True,
                        )
                        if cur_size == total_size:
                            print("\n")
                self.inputs[input_no] = os.path.abspath(fpath)
            except:
                print(f"File downloading unsuccessful. Exiting.")
                exit()
            return None
        else:
            return ip

    def get_logger(self):
        from ..exceptions import SetupError
        import os
        import logging

        if (
            self.args is None
            or self.run_name is None
            or self.output_dir is None
        ):
            raise SetupError()
        if self.args.newlog == True:
            self.logmode = "w"
        else:
            self.logmode = "a"
        self.logger = logging.getLogger("oakvar")
        self.logger.setLevel("INFO")
        self.log_path = os.path.join(self.output_dir, self.run_name + ".log")
        if os.path.exists(self.log_path):
            os.remove(self.log_path)
        self.log_handler = logging.FileHandler(self.log_path, mode=self.logmode)
        formatter = logging.Formatter(
            "%(asctime)s %(name)-20s %(message)s", "%Y/%m/%d %H:%M:%S"
        )
        self.log_handler.setFormatter(formatter)
        self.logger.addHandler(self.log_handler)
        # individual input line error log
        self.error_logger = logging.getLogger("err")
        self.error_logger.setLevel("INFO")
        error_log_path = os.path.join(self.output_dir, self.run_name + ".err")
        if os.path.exists(error_log_path):
            os.remove(error_log_path)
        self.error_log_handler = logging.FileHandler(error_log_path, mode=self.logmode)
        formatter = logging.Formatter("%(name)s\t%(message)s")
        self.error_log_handler.setFormatter(formatter)
        self.error_logger.addHandler(self.error_log_handler)

    def log_versions(self):
        from ..exceptions import SetupError
        import os
        from ..util import admin_util as au
        from ..exceptions import ModuleLoadingError

        if self.args is None:
            raise SetupError()
        if self.logger:
            self.logger.info(
                f"version: oakvar=={au.get_current_package_version()} {au.get_packagedir()}"
            )
            if self.package_conf is not None and len(self.package_conf) > 0:
                self.logger.info(
                    f'package: {self.args.package} {self.package_conf.get("version")}'
                )
            for _, module in self.annotators.items():
                if module.conf:
                    version = module.conf.get("code_version")
                    if not version:
                        version = module.conf.get("version")
                    self.logger.info(
                        f"version: {module.name}=={version} {os.path.dirname(module.script_path)}"
                    )
            if "mapper" not in self.args.skip:
                module = self.mapper
                if module is None:
                    raise ModuleLoadingError("mapper")
                if module.conf:
                    version = module.conf.get("code_version")
                    if not version:
                        version = module.conf.get("version")
                    self.logger.info(
                        f"version: {module.name} {version} {os.path.dirname(module.script_path)}"
                    )
            for _, module in self.reports.items():
                if module.conf:
                    version = module.conf.get("code_version")
                    if not version:
                        version = module.conf.get("version")
                    self.logger.info(
                        f"version: {module.name} {version} {os.path.dirname(module.script_path)}"
                    )

    async def do_step_converter(self):
        from ..util.util import quiet_print
        from time import time

        if (
            self.endlevel >= self.runlevels["converter"]
            and self.startlevel <= self.runlevels["converter"]
            and (self.args and not "converter" in self.args.skip)
        ):
            quiet_print("Running converter...", self.args)
            stime = time()
            self.run_converter()
            rtime = time() - stime
            quiet_print("finished in {0:.3f}s".format(rtime), self.args)
            if self.numinput == 0:
                msg = "No variant found in input"
                quiet_print(msg, self.args)
                if self.logger:
                    self.logger.info(msg)
                exit()

    async def do_step_mapper(self):
        from ..util.util import quiet_print
        from time import time

        self.mapper_ran = False
        if (
            self.endlevel >= self.runlevels.get("mapper", 0)
            and self.startlevel <= self.runlevels.get("mapper", 0)
            and (self.args and not "mapper" in self.args.skip)
        ):
            quiet_print(f'Running gene mapper...{" "*18}', self.args)
            stime = time()
            # multicore_mapper_mode = self.main_conf.get("multicore_mapper_mode")
            self.run_genemapper_mp()
            # if multicore_mapper_mode:
            #    self.run_genemapper_mp()
            # else:
            #    self.run_genemapper()
            rtime = time() - stime
            quiet_print("finished in {0:.3f}s".format(rtime), self.args)
            self.mapper_ran = True

    async def do_step_annotator(self):
        from ..util.util import quiet_print
        from time import time

        self.annotator_ran = False
        self.done_annotators = {}
        self.populate_secondary_annotators()
        for mname, module in self.annotators.items():
            if self.check_module_output(module) is not None:
                self.done_annotators[mname] = module
        self.run_annotators = {
            aname: self.annotators[aname]
            for aname in set(self.annotators) - set(self.done_annotators)
        }
        if (
            self.endlevel >= self.runlevels["annotator"]
            and self.startlevel <= self.runlevels["annotator"]
            and (self.args and not "annotator" in self.args.skip)
            and (self.mapper_ran or len(self.run_annotators) > 0)
        ):
            quiet_print("Running annotators...", self.args)
            stime = time()
            self.run_annotators_mp()
            rtime = time() - stime
            quiet_print("\tannotator(s) finished in {0:.3f}s".format(rtime), self.args)

    async def do_step_aggregator(self):
        from ..util.util import quiet_print

        if (
            self.endlevel >= self.runlevels["aggregator"]
            and self.startlevel <= self.runlevels["aggregator"]
            and (self.args and not "aggregator" in self.args.skip)
            and (
                self.mapper_ran
                or self.annotator_ran
                or self.startlevel == self.runlevels["aggregator"]
            )
        ):
            quiet_print("Running aggregator...", self.args)
            self.result_path = self.run_aggregator()
            await self.write_job_info()
            self.write_smartfilters()
            self.aggregator_ran = True

    async def do_step_postaggregator(self):
        from ..util.util import quiet_print

        if (
            self.endlevel >= self.runlevels["postaggregator"]
            and self.startlevel <= self.runlevels["postaggregator"]
            and (self.args and not "postaggregator" in self.args.skip)
        ):
            quiet_print("Running postaggregators...", self.args)
            self.run_postaggregators()

    async def do_step_reporter(self):
        from ..util.util import quiet_print

        if (
            self.endlevel >= self.runlevels["reporter"]
            and self.startlevel <= self.runlevels["reporter"]
            and (self.args and not "reporter" in self.args.skip)
            and self.aggregator_ran
            and self.reports
        ):
            quiet_print("Running reporter...", self.args)
            self.report_response = await self.run_reporter()

    async def setup_manager(self):
        from multiprocessing.managers import SyncManager

        self.manager = SyncManager()
        self.manager.register("StatusWriter", StatusWriter)
        self.manager.start()
        self.status_writer = self.manager.StatusWriter(  # type: ignore
            self.status_json_path
        )

    async def process_clean(self):
        from ..util.util import quiet_print

        if self.args and self.args.clean:
            quiet_print("Deleting previous output files...", self.args)
            self.delete_output_files()

    def process_input(self):
        from ..exceptions import NoInput

        if self.inputs is None:
            raise NoInput()
        if self.pipeinput == False:
            input_files_str = ", ".join(self.inputs)
        else:
            input_files_str = "stdin"
        if self.logger:
            self.logger.info("input files: {}".format(input_files_str))

    async def start_logger(self):
        from time import asctime, localtime
        from sys import argv

        self.unique_logs = {}
        self.get_logger()
        if self.logger:
            self.logger.info(f'{" ".join(argv)}')
            self.logger.info("started: {0}".format(asctime(localtime(self.start_time))))
            if self.conf_path != "":
                self.logger.info("conf file: {}".format(self.conf_path))

    async def main(self):
        from ..util.util import quiet_print
        from ..util.util import update_status
        from time import time, asctime, localtime

        self.report_response = None
        self.aggregator_ran = False
        try:
            self.start_time = time()
            self.make_args_namespace(self.inkwargs)
            await self.process_clean()
            await self.start_logger()
            self.write_initial_status_json()
            await self.setup_manager()
            update_status(
                "Started OakVar",
                status_writer=self.status_writer,
                args=self.args,
                force=True,
            )
            self.process_input()
            self.log_versions()
            self.set_and_check_input_files()
            if self.args and self.args.vcf2vcf:
                await self.run_vcf2vcf()
            else:
                await self.do_step_converter()
                await self.do_step_mapper()
                await self.do_step_annotator()
                await self.do_step_aggregator()
                await self.do_step_postaggregator()
                await self.do_step_reporter()
            update_status(
                "Finished", status_writer=self.status_writer, args=self.args, force=True
            )
        except Exception as e:
            self.exception = e
            # self.handle_exception(e)
        finally:
            end_time = time()
            display_time = asctime(localtime(end_time))
            self.logger = getattr(self, "logger", None)
            self.status_writer = getattr(self, "status_writer", None)
            runtime = None
            if self.start_time:
                runtime = end_time - self.start_time
            if not self.exception:
                if self.logger:
                    self.logger.info("finished: {0}".format(display_time))
                    if runtime:
                        self.logger.info("runtime: {0:0.3f}s".format(runtime))
                    quiet_print(
                        "Finished normally. Runtime: {0:0.3f}s".format(runtime),
                        self.args,
                    )
            else:
                # if self.args and not self.args.quiet:
                # if isinstance(exception, ExpectedException):
                # if self.logger and self.log_path:
                # from sys import stderr
                # stderr.write(
                #    "Log file at {}".format(self.log_path) + "\n")
                # stderr.flush()
                if self.status_writer:
                    update_status(
                        "Error",
                        status_writer=self.status_writer,
                        args=self.args,
                        force=True,
                    )
            if self.logger:
                self.close_logger()
            if (
                self.status_writer
                and self.status_writer
                and self.args
                and self.args.do_not_change_status != True
            ):
                self.status_writer.flush()
            if (
                not self.exception
                and self.args
                and not self.args.temp_files
                and self.aggregator_ran
            ):
                self.clean_up_at_end()
            if self.args and self.args.writeadmindb:
                await self.write_admin_db(runtime, self.numinput)
            if self.exception:
                raise self.exception
            return self.report_response

    def make_args_namespace(self, args):
        from ..exceptions import SetupError
        from ..exceptions import SetupError
        from .version import cli_version
        from ..exceptions import NormalExit

        self.set_package_conf(args)
        self.make_self_args_considering_package_conf(args)
        if self.args is None:
            raise SetupError()
        if self.args.show_version:
            cli_version({"to": "stdout"})
            raise NormalExit
        self.set_self_inputs()
        self.set_output_dir()
        self.set_run_name()
        self.set_append_mode()
        if self.args.skip is None:
            self.args.skip = []
        self.set_md()
        self.set_mapper()
        self.set_annotators()
        self.set_postaggregators()
        self.set_reporters()
        self.verbose = self.args.verbose == True
        self.set_start_end_levels()
        self.cleandb = self.args.cleandb
        if self.args.note == None:
            self.args.note = ""
        if self.args is None:
            raise SetupError("Cravat")

    def make_self_args_considering_package_conf(self, args):
        from types import SimpleNamespace
        from ..util.admin_util import get_user_conf

        full_args = args
        # package including -a (add) and -A (replace)
        if "run" in self.package_conf:
            for k, v in self.package_conf.get("run", {}).items():
                if k == "annotators" and v and isinstance(v, list):
                    if not full_args.get("annotators_replace"):
                        for v2 in v:
                            if v2 not in full_args.get("annotators", []):
                                full_args["annotators"].append(v2)
                else:
                    if k not in full_args or not full_args[k]:
                        full_args[k] = v
        self.conf_path = full_args.get("confpath", None)
        self.make_self_conf(full_args)
        self.main_conf = get_user_conf() or {}
        self.conf_run = self.conf.get("run", {})
        for k, v in self.conf_run.items():
            if k not in full_args or (not full_args[k] and v):
                full_args[k] = v
        if full_args.get("annotators_replace"):
            full_args["annotators"] = full_args.get("annotators_replace")

        self.args = SimpleNamespace(**full_args)
        self.process_module_options()

    def make_self_conf(self, args):
        from ..exceptions import SetupError
        from ..util.util import quiet_print
        import json

        if args is None:
            raise SetupError()
        self.conf_run = args.get("conf", {}).get("run", {})
        confs = args.get("confs")
        if confs:
            conf_bak = self.conf
            try:
                confs_conf = json.loads(confs.replace("'", '"'))
                self.conf.update(confs_conf)
            except Exception:
                quiet_print(
                    "Error in processing cs option. --cs option was not applied.",
                    self.args,
                )
                self.conf = conf_bak

    def populate_secondary_annotators(self):
        import os

        secondaries = {}
        for module in self.annotators.values():
            self._find_secondary_annotators(module, secondaries)
        self.annotators.update(secondaries)
        annot_names = [v.name for v in self.annotators.values()]
        annot_names = list(set(annot_names))
        filenames = os.listdir(self.output_dir)
        for filename in filenames:
            toks = filename.split(".")
            if len(toks) == 3:
                extension = toks[2]
                if toks[0] == self.run_name and (
                    extension == "var" or extension == "gen"
                ):
                    annot_name = toks[1]
                    if annot_name not in annot_names:
                        annot_names.append(annot_name)
        annot_names.sort()
        if (
            self.status_writer is not None
            and self.startlevel <= self.runlevels["annotator"]
        ):
            self.status_writer.queue_status_update(
                "annotators", annot_names, force=True
            )
        self.annot_names = annot_names

    def process_module_options(self):
        from ..exceptions import SetupError
        from ..util.util import quiet_print

        if self.args is None or self.conf is None:
            raise SetupError()
        if self.args.module_option is not None:
            for opt_str in self.args.module_option:
                toks = opt_str.split("=")
                if len(toks) != 2:
                    quiet_print(
                        "Ignoring invalid module option {opt_str}. module-option should be module_name.key=value.",
                        self.args,
                    )
                    continue
                k = toks[0]
                if k.count(".") != 1:
                    quiet_print(
                        "Ignoring invalid module option {opt_str}. module-option should be module_name.key=value.",
                        self.args,
                    )
                    continue
                [module_name, key] = k.split(".")
                if module_name not in self.conf_run:
                    self.conf_run[module_name] = {}
                v = toks[1]
                self.conf_run[module_name][key] = v

    def process_url_and_pipe_inputs(self):
        from ..exceptions import SetupError
        from ..util.util import is_url
        from ..exceptions import InvalidInputFormat
        from ..exceptions import NoInput
        import os

        if self.args is None:
            raise SetupError()
        self.first_non_url_input = None
        if (
            self.args.inputs is not None
            and len(self.args.inputs) == 1
            and self.args.inputs[0] == "-"
        ):
            self.pipeinput = True
            if self.args.forcedinputformat is None:
                raise InvalidInputFormat(fmt="--input-format is needed for pipe input.")
        if self.args.inputs is not None:
            self.inputs = [
                os.path.abspath(x) if not is_url(x) and x != "-" else x
                for x in self.args.inputs
            ]
            if self.inputs is None:
                raise NoInput()
            for input_no in range(len(self.inputs)):
                if (
                    self.download_url_input(input_no) is not None
                    and self.first_non_url_input is None
                ):
                    self.first_non_url_input = self.inputs[input_no]
        else:
            self.inputs = []

    def regenerate_from_db(self):
        from ..exceptions import NoInput
        import sqlite3
        from ..consts import crv_def, crx_def, crg_def
        from ..util.inout import CravatWriter

        if self.inputs is None or len(self.inputs) == 0:
            raise NoInput
        dbpath = self.inputs[0]
        db = sqlite3.connect(dbpath)
        c = db.cursor()
        # Variant
        if not self.crv_present:
            crv = CravatWriter(self.crvinput, columns=crv_def)
            crv.write_definition()
        else:
            crv = None
        if not self.crx_present:
            crx = CravatWriter(self.crxinput, columns=crx_def)
            crx.write_definition()
        else:
            crx = None
        if crv or crx:
            colnames = [x["name"] for x in crx_def]
            sel_cols = ", ".join(["base__" + x for x in colnames])
            q = f"select {sel_cols} from variant"
            c.execute(q)
            for r in c:
                rd = {x[0]: x[1] for x in zip(colnames, r)}
                if crv:
                    crv.write_data(rd)
                if crx:
                    crx.write_data(rd)
            if crv:
                crv.close()
            if crx:
                crx.close()
            self.crv_present = True
            self.crx_present = True
        # Gene
        if not self.crg_present:
            crg = CravatWriter(self.crginput, columns=crg_def)
            crg.write_definition()
            colnames = [x["name"] for x in crg_def]
            sel_cols = ", ".join(["base__" + x for x in colnames])
            q = f"select {sel_cols} from gene"
            c.execute(q)
            for r in c:
                rd = {x[0]: x[1] for x in zip(colnames, r)}
                crg.write_data(rd)
            crg.close()
            self.crg_present = True
        c.close()
        db.close()

    def set_append_mode(self):
        import os
        import shutil
        from ..exceptions import NoInput
        from ..exceptions import SetupError

        if self.inputs is None or self.num_input is None:
            raise NoInput()
        if self.run_name is None or self.args is None:
            raise SetupError()
        if self.num_input > 0 and self.inputs[0].endswith(".sqlite"):
            self.append_mode = True
            if self.args.skip is None:
                self.args.skip = ["converter", "mapper"]
            else:
                if "converter" not in self.args.skip:
                    self.args.skip.append("converter")
                if "mapper" not in self.args.skip:
                    self.args.skip.append("mapper")
            if self.args.output_dir:
                if self.run_name.endswith(".sqlite"):
                    target_name = self.run_name
                else:
                    target_name = self.run_name + ".sqlite"
                target_path = os.path.join(self.args.output_dir, target_name)
                shutil.copyfile(self.inputs[0], target_path)
                self.inputs[0] = target_path
            if self.run_name.endswith(".sqlite"):
                self.run_name = self.run_name[:-7]

    def set_output_dir(self):
        import os
        from ..exceptions import SetupError

        if self.args is None:
            raise SetupError()
        self.output_dir = self.args.output_dir
        if self.output_dir == None:
            if self.num_input == 0 or self.first_non_url_input is None:
                self.output_dir = os.getcwd()
            else:
                self.output_dir = os.path.dirname(
                    os.path.abspath(self.first_non_url_input)
                )
        else:
            self.output_dir = os.path.abspath(self.output_dir)
        if os.path.exists(self.output_dir) == False:
            os.mkdir(self.output_dir)

    def set_package_conf(self, args):
        from ..module.cache import get_module_cache

        package_name = args.get("package", None)
        if package_name:
            if package_name in get_module_cache().get_local():
                self.package_conf = get_module_cache().get_local()[package_name].conf
            else:
                self.package_conf = {}
        else:
            self.package_conf = {}

    def set_run_name(self):
        import os
        from ..exceptions import NoInput
        from ..exceptions import SetupError
        from ..exceptions import ArgumentError

        if self.inputs is None or self.num_input is None:
            raise NoInput()
        if self.args is None:
            raise SetupError()
        self.run_name = self.args.run_name
        if self.run_name == None:
            if self.num_input == 0 or self.pipeinput:
                self.run_name = "ovjob"
            else:
                self.run_name = os.path.basename(self.inputs[0])
                if self.num_input > 1:
                    e = ArgumentError("--run_name should be given when multiple input files are given.")
                    e.traceback = False
                    e.halt = True
                    raise e

    def set_self_inputs(self):
        from ..exceptions import SetupError
        from ..util.util import quiet_print
        from ..exceptions import NoInput

        if self.args is None:
            raise SetupError()
        if self.args.inputs and len(self.args.inputs) == 0:
            inputs = self.conf_run.get("inputs")
            if inputs:
                if type(inputs) == list:
                    self.args.inputs = inputs
                else:
                    quiet_print("inputs in conf file is invalid", self.args)
            else:
                raise NoInput()
        self.process_url_and_pipe_inputs()
        if self.inputs is None:
            raise NoInput()
        self.num_input = len(self.inputs)

    def set_start_end_levels(self):
        from ..exceptions import SetupError

        if self.args is None:
            raise SetupError()
        self.startlevel = self.runlevels.get(self.args.startat, 0)
        if self.append_mode:
            if self.args.endat is None:
                if len(self.report_names) > 0:
                    self.args.endat = "reporter"
                else:
                    self.args.endat = "aggregator"
        self.endlevel = self.runlevels.get(
            self.args.endat, max(self.runlevels.values())
        )

    def set_and_check_input_files(self):
        from ..exceptions import SetupError
        from ..exceptions import NoInput
        import os

        if self.run_name is None or self.output_dir is None:
            raise SetupError()
        if self.inputs is None or len(self.inputs) == 0:
            raise NoInput
        self.crvinput = os.path.join(self.output_dir, self.run_name + ".crv")
        self.crxinput = os.path.join(self.output_dir, self.run_name + ".crx")
        self.crginput = os.path.join(self.output_dir, self.run_name + ".crg")
        if os.path.exists(self.crvinput):
            self.crv_present = True
        else:
            self.crv_present = False
        if os.path.exists(self.crxinput):
            self.crx_present = True
        else:
            self.crx_present = False
        if os.path.exists(self.crginput):
            self.crg_present = True
        else:
            self.crg_present = False
        if self.append_mode:
            self.regenerate_from_db()
        return True

    def set_annotators(self):
        from ..exceptions import SetupError
        from ..module.local import get_local_module_infos_of_type
        from ..module.local import get_local_module_infos_by_names

        if self.args is None:
            raise SetupError()
        self.excludes = self.args.excludes
        if len(self.args.annotators) > 0:
            if self.args.annotators == ["all"]:
                self.annotator_names = sorted(
                    list(get_local_module_infos_of_type("annotator").keys())
                )
            else:
                self.annotator_names = self.args.annotators
        elif (
            self.package_conf is not None
            and "run" in self.package_conf
            and "annotators" in self.package_conf.get("run", {})
        ):
            self.annotator_names = self.package_conf.get("run", {}).get("annotators")
        else:
            self.annotator_names = []
        if "annotator" in self.args.skip:
            self.annotator_names = []
        elif len(self.excludes) > 0:
            if "all" in self.excludes:
                self.annotator_names = []
            else:
                for m in self.excludes:
                    if self.annotator_names and m in self.annotator_names:
                        self.annotator_names.remove(m)
        self.check_valid_modules(self.annotator_names)
        self.annotators = get_local_module_infos_by_names(self.annotator_names)

    def set_mapper(self):
        from ..module.local import get_local_module_info_by_name
        from ..exceptions import SetupError

        if self.args is None or self.conf is None:
            raise SetupError()
        if self.args.mapper_name:
            self.mapper_name = self.args.mapper_name[0]
        self.mapper_name = self.package_conf.get("run", {}).get("mapper")
        if not self.mapper_name:
            self.mapper_name = self.main_conf.get("genemapper")
        self.check_valid_modules([self.mapper_name])
        self.mapper = get_local_module_info_by_name(self.mapper_name)

    def set_md(self):
        from ..exceptions import SetupError
        from ..system import consts

        if self.args is None:
            raise SetupError()
        if self.args.md is not None:
            consts.custom_modules_dir = self.args.md

    def set_postaggregators(self):
        from ..exceptions import SetupError
        from ..system.consts import default_postaggregator_names
        from ..module.local import get_local_module_infos_by_names
        from ..module.local import module_exists_local

        if self.args is None:
            raise SetupError()
        if len(self.args.postaggregators) > 0:
            self.postaggregator_names = self.args.postaggregators
        elif (
            self.package_conf is not None
            and self.package_conf.get("run")
            and self.package_conf["run"].get("postaggregators")
        ):
            self.postaggregator_names = sorted(
                list(
                    get_local_module_infos_by_names(
                        self.package_conf["run"]["postaggregators"]
                    )
                )
            )
        else:
            self.postaggregator_names = []
        if "postaggregator" in self.args.skip:
            self.postaggregators = {}
        else:
            self.postaggregator_names = sorted(
                list(
                    set(self.postaggregator_names).union(
                        set(default_postaggregator_names)
                    )
                )
            )
            if "casecontrol" in self.postaggregator_names:
                if module_exists_local("casecontrol") == False:
                    self.postaggregator_names.remove("casecontrol")
            self.check_valid_modules(self.postaggregator_names)
            self.postaggregators = get_local_module_infos_by_names(
                self.postaggregator_names
            )

    def set_reporters(self):
        from ..module.local import get_local_module_infos_by_names
        from ..exceptions import SetupError

        if self.args is None:
            raise SetupError()
        if len(self.args.reports) > 0:
            self.report_names = self.args.reports
        elif (
            self.package_conf is not None
            and self.package_conf.get("run")
            and self.package_conf["run"].get("reports")
        ):
            self.report_names = self.package_conf["run"]["reports"]
        else:
            self.report_names = []
        if "reporter" in self.args.skip:
            self.reports = {}
        else:
            self.reporter_names = [v + "reporter" for v in self.report_names]
            self.check_valid_modules(self.reporter_names)
            self.reports = get_local_module_infos_by_names(self.reporter_names)

    def _find_secondary_annotators(self, module, ret):
        sannots = self.get_secondary_modules(module)
        for sannot in sannots:
            if sannot is not None:
                ret[sannot.name] = sannot
                self._find_secondary_annotators(sannot, ret)

    def get_module_output_path(self, module):
        from ..exceptions import SetupError
        import os

        if not self.run_name or not self.output_dir:
            raise SetupError()
        if module.level == "variant":
            postfix = ".var"
        elif module.level == "gene":
            postfix = ".gen"
        else:
            return None
        path = os.path.join(
            self.output_dir, self.run_name + "." + module.name + postfix
        )
        return path

    def check_module_output(self, module):
        import os

        path = self.get_module_output_path(module)
        if path is not None and os.path.exists(path):
            return path
        else:
            return None

    def get_secondary_modules(self, primary_module):
        from ..module.local import get_local_module_info

        secondary_modules = [
            get_local_module_info(module_name)
            for module_name in primary_module.secondary_module_names
        ]
        return secondary_modules

    def run_converter(self):
        import os
        from ..util.util import load_class
        from types import SimpleNamespace
        from ..exceptions import SetupError
        from ..util.admin_util import get_packagedir
        from ..util.util import announce_module

        if self.conf is None or self.args is None:
            raise SetupError()
        converter_path = os.path.join(get_packagedir(), "base", "cravat_convert.py")
        module = SimpleNamespace(
            title="Converter", name="converter", script_path=converter_path
        )
        arg_dict = {
            "path": module.script_path,
            "inputs": self.inputs,
            "name": self.run_name,
            "output_dir": self.output_dir,
            "genome": self.args.genome,
        }
        arg_dict["conf"] = self.conf_run
        if self.args.forcedinputformat is not None:
            arg_dict["format"] = self.args.forcedinputformat
        if self.args.unique_variants:
            arg_dict["unique_variants"] = True
        announce_module(module, status_writer=self.status_writer, args=self.args)
        if self.verbose:
            print(
                " ".join([str(k) + "=" + str(v) for k, v in arg_dict.items()]),
                self.args,
            )
        arg_dict["status_writer"] = self.status_writer
        converter_class = load_class(module.script_path, "MasterCravatConverter")
        converter = converter_class(arg_dict)
        self.numinput, self.converter_format, self.genome_assembiles = converter.run()

    def run_genemapper(self):
        from ..module.local import get_local_module_info
        from ..exceptions import ModuleLoadingError
        from ..util.util import load_class
        from ..util.util import quiet_print
        from ..exceptions import SetupError

        if self.args is None:
            raise SetupError()
        module = get_local_module_info(self.main_conf.get("genemapper"))
        if module is None:
            raise ModuleLoadingError(self.main_conf.get("genemapper"))
        self.genemapper = module
        cmd = [
            module.script_path,
            self.crvinput,
            "-n",
            self.run_name,
            "-d",
            self.output_dir,
        ]
        if self.args.primary_transcript is not None:
            if "mane" not in self.args.primary_transcript:
                self.args.primary_transcript.append("mane")
            cmd.extend(["--primary-transcript"])
            cmd.extend(self.args.primary_transcript)
        if self.verbose:
            quiet_print(" ".join(cmd), self.args)
        genemapper_class = load_class(module.script_path, "Mapper")
        genemapper = genemapper_class(cmd, self.status_writer)
        genemapper.run()

    def get_num_workers(self) -> int:
        from ..system import get_max_num_concurrent_annotators_per_job
        from psutil import cpu_count

        num_workers = get_max_num_concurrent_annotators_per_job()
        if self.args and self.args.mp:
            try:
                self.args.mp = int(self.args.mp)
                if self.args.mp >= 1:
                    num_workers = self.args.mp
            except:
                if self.logger:
                    self.logger.exception(
                        f"error handling --mp argument: {self.args.mp}"
                    )
        if not num_workers:
            num_workers = cpu_count()
        if self.logger:
            self.logger.info("num_workers: {}".format(num_workers))
        return num_workers

    def collect_crxs(self):
        from os.path import join
        from glob import glob
        from os import remove

        if self.output_dir:
            crx_path = join(self.output_dir, f"{self.run_name}.crx")
            wf = open(crx_path, "w")
            fns = sorted(glob(crx_path + "[.]*"))
            fn = fns[0]
            f = open(fn)
            for line in f:
                wf.write(line)
            f.close()
            remove(fn)
            for fn in fns[1:]:
                f = open(fn)
                for line in f:
                    if line[0] != "#":
                        wf.write(line)
                f.close()
                remove(fn)
            wf.close()

    def collect_crgs(self):
        from os.path import join
        from glob import glob
        from os import remove

        if self.output_dir:
            crg_path = join(self.output_dir, f"{self.run_name}.crg")
            wf = open(crg_path, "w")
            unique_hugos = {}
            fns = sorted(glob(crg_path + "[.]*"))
            fn = fns[0]
            f = open(fn)
            for line in f:
                if line[0] != "#":
                    hugo = line.split()[0]
                    if hugo not in unique_hugos:
                        # wf.write(line)
                        unique_hugos[hugo] = line
                else:
                    wf.write(line)
            f.close()
            remove(fn)
            for fn in fns[1:]:
                f = open(fn)
                for line in f:
                    if line[0] != "#":
                        hugo = line.split()[0]
                        if hugo not in unique_hugos:
                            # wf.write(line)
                            unique_hugos[hugo] = line
                f.close()
                remove(fn)
            hugos = list(unique_hugos.keys())
            hugos.sort()
            for hugo in hugos:
                wf.write(unique_hugos[hugo])
            wf.close()
            del unique_hugos
            del hugos

    def run_genemapper_mp(self):
        import multiprocessing as mp
        from ..base.mp_runners import init_worker, mapper_runner
        from ..util.inout import CravatReader
        from ..exceptions import SetupError

        if self.args is None or self.output_dir is None:
            raise SetupError()
        num_workers = self.get_num_workers()
        reader = CravatReader(self.crvinput)
        num_lines, chunksize, poss, len_poss, max_num_lines = reader.get_chunksize(
            num_workers
        )
        if self.logger:
            self.logger.info(
                f"input line chunksize={chunksize} total number of input lines={num_lines} number of chunks={len_poss}"
            )
        pool = mp.Pool(num_workers, init_worker)
        pos_no = 0
        while pos_no < len_poss:
            jobs = []
            for _ in range(num_workers):
                if pos_no == len_poss:
                    break
                (seekpos, num_lines) = poss[pos_no]
                if pos_no == len_poss - 1:
                    job = pool.apply_async(
                        mapper_runner,
                        (
                            self.crvinput,
                            seekpos,
                            max_num_lines - num_lines,
                            self.run_name,
                            self.output_dir,
                            self.status_writer,
                            self.mapper_name,
                            pos_no,
                            ";".join(self.args.primary_transcript),
                        ),
                    )
                else:
                    job = pool.apply_async(
                        mapper_runner,
                        (
                            self.crvinput,
                            seekpos,
                            chunksize,
                            self.run_name,
                            self.output_dir,
                            self.status_writer,
                            self.mapper_name,
                            pos_no,
                            ";".join(self.args.primary_transcript),
                        ),
                    )
                jobs.append(job)
                pos_no += 1
            for job in jobs:
                job.get()
        pool.close()
        self.collect_crxs()
        self.collect_crgs()

    def run_aggregator(self):
        from time import time
        from ..base.aggregator import Aggregator
        from ..util.util import quiet_print
        from ..util.util import update_status

        # Variant level
        quiet_print("\t{0:30s}\t".format("Variants"), self.args)
        stime = time()
        cmd = [
            "donotremove",
            "-i",
            self.output_dir,
            "-d",
            self.output_dir,
            "-l",
            "variant",
            "-n",
            self.run_name,
        ]
        if self.cleandb:
            cmd.append("-x")
        if self.append_mode:
            cmd.append("--append")
        if self.verbose:
            quiet_print(" ".join(cmd), self.args)
        update_status(
            "Running {title} ({level})".format(title="Aggregator", level="variant"),
            status_writer=self.status_writer,
            args=self.args,
            force=True,
        )
        v_aggregator = Aggregator(cmd, self.status_writer)
        v_aggregator.run()
        rtime = time() - stime
        quiet_print("finished in {0:.3f}s".format(rtime), self.args)
        # Gene level
        quiet_print("\t{0:30s}\t".format("Genes"), self.args)
        stime = time()
        cmd = [
            "donotremove",
            "-i",
            self.output_dir,
            "-d",
            self.output_dir,
            "-l",
            "gene",
            "-n",
            self.run_name,
        ]
        if self.append_mode:
            cmd.append("--append")
        if self.verbose:
            quiet_print(" ".join(cmd), self.args)
        update_status(
            "Running {title} ({level})".format(title="Aggregator", level="gene"),
            status_writer=self.status_writer,
            args=self.args,
            force=True,
        )
        g_aggregator = Aggregator(cmd, self.status_writer)
        g_aggregator.run()
        rtime = time() - stime
        quiet_print("finished in {0:.3f}s".format(rtime), self.args)
        # Sample level
        if not self.append_mode:
            quiet_print("\t{0:30s}\t".format("Samples"), self.args)
            stime = time()
            cmd = [
                "donotremove",
                "-i",
                self.output_dir,
                "-d",
                self.output_dir,
                "-l",
                "sample",
                "-n",
                self.run_name,
            ]
            if self.verbose:
                quiet_print(" ".join(cmd), self.args)
            update_status(
                "Running {title} ({level})".format(title="Aggregator", level="sample"),
                status_writer=self.status_writer,
                args=self.args,
                force=True,
            )
            s_aggregator = Aggregator(cmd, self.status_writer)
            s_aggregator.run()
            rtime = time() - stime
            quiet_print("finished in {0:.3f}s".format(rtime), self.args)
        # Mapping level
        if not self.append_mode:
            quiet_print("\t{0:30s}\t".format("Tags"), self.args)
            cmd = [
                "donotremove",
                "-i",
                self.output_dir,
                "-d",
                self.output_dir,
                "-l",
                "mapping",
                "-n",
                self.run_name,
            ]
            if self.verbose:
                quiet_print(" ".join(cmd), self.args)
            update_status(
                "Running {title} ({level})".format(title="Aggregator", level="mapping"),
                status_writer=self.status_writer,
                args=self.args,
                force=True,
            )
            m_aggregator = Aggregator(cmd, self.status_writer)
            m_aggregator.run()
            rtime = time() - stime
            quiet_print("finished in {0:.3f}s".format(rtime), self.args)
        return v_aggregator.db_path

    def run_postaggregators(self):
        from ..util.util import announce_module
        from ..exceptions import SetupError
        from time import time
        from ..util.util import load_class
        import json
        from ..util.util import quiet_print
        from ..system.consts import default_postaggregator_names

        if self.conf is None:
            raise SetupError()
        for module_name, module in self.postaggregators.items():
            if self.append_mode and module_name in default_postaggregator_names:
                continue
            cmd = [module.script_path, "-d", self.output_dir, "-n", self.run_name]
            postagg_conf = {}
            postagg_conf.update(self.conf_run.get(module_name, {}))
            if postagg_conf:
                confs = json.dumps(postagg_conf)
                confs = "'" + confs.replace("'", '"') + "'"
                cmd.extend(["--confs", confs])
            if self.verbose:
                quiet_print(" ".join(cmd), self.args)
            post_agg_cls = load_class(module.script_path, "CravatPostAggregator")
            post_agg = post_agg_cls(cmd, self.status_writer)
            if post_agg.should_run_annotate:
                announce_module(
                    module, status_writer=self.status_writer, args=self.args
                )
            stime = time()
            post_agg.run()
            rtime = time() - stime
            if post_agg.should_run_annotate:
                quiet_print("finished in {0:.3f}s".format(rtime), self.args)

    async def run_vcf2vcf(self):
        from ..exceptions import SetupError
        from ..exceptions import NoInput
        from time import time
        from ..util.util import load_class
        from ..util.util import quiet_print
        from types import SimpleNamespace
        from ..base import vcf2vcf
        from os.path import abspath

        if self.conf is None or self.args is None or self.output_dir is None:
            raise SetupError()
        if self.inputs is None:
            raise NoInput()
        response = {}
        module = {
            "name": "vcf2vcf",
            "title": "VCF to VCF",
            "script_path": abspath(vcf2vcf.__file__),
        }
        module = SimpleNamespace(**module)
        # announce_module(module, status_writer=self.status_writer, args=self.args)
        arg_dict = dict(vars(self.args))
        arg_dict["output_dir"] = self.output_dir
        arg_dict["module_name"] = module.name
        arg_dict["conf"] = self.conf
        arg_dict["mapper_name"] = self.mapper_name
        arg_dict["annotator_names"] = self.annotator_names
        arg_dict["output_dir"] = self.output_dir
        arg_dict["run_name"] = self.run_name
        Module = load_class(module.script_path, "VCF2VCF")
        m = Module(arg_dict)
        stime = time()
        response_t = m.run()
        output_fns = None
        response_type = type(response_t)
        if response_type == list:
            output_fns = " ".join(response_t)
        elif response_type == str:
            output_fns = response_t
        if output_fns is not None:
            quiet_print(f"report created: {output_fns} ", self.args)
        report_type = "vcf2vcf"
        response[report_type] = response_t
        rtime = time() - stime
        quiet_print("finished in {0:.3f}s".format(rtime), self.args)
        self.report_response = response

    async def run_reporter(self):
        from ..module.local import get_local_module_info
        from ..exceptions import SetupError
        from ..exceptions import NoInput
        from time import time
        import os
        from ..util.util import load_class
        from ..util.util import quiet_print
        from ..exceptions import ModuleNotExist
        from ..util.util import announce_module

        if (
            self.run_name is None
            or self.conf is None
            or self.args is None
            or self.output_dir is None
        ):
            raise SetupError()
        if self.inputs is None:
            raise NoInput()
        if len(self.reports) > 0:
            module_names = [v for v in self.reports.keys()]
            report_types = [v.replace("reporter", "") for v in self.reports.keys()]
        else:
            module_names = []
            report_types = []
        response = {}
        for report_type, module_name in zip(report_types, module_names):
            reporter = None
            module = get_local_module_info(module_name)
            announce_module(module, status_writer=self.status_writer, args=self.args)
            if module is None:
                raise ModuleNotExist(module_name)
            arg_dict = dict(vars(self.args))
            arg_dict["script_path"] = module.script_path
            arg_dict["dbpath"] = os.path.join(
                self.output_dir, self.run_name + ".sqlite"
            )
            arg_dict["savepath"] = os.path.join(self.output_dir, self.run_name)
            arg_dict["output_dir"] = self.output_dir
            arg_dict["module_name"] = module_name
            arg_dict["conf"] = self.conf
            Reporter = load_class(module.script_path, "Reporter")
            reporter = Reporter(arg_dict)
            await reporter.prep()
            stime = time()
            response_t = await reporter.run()
            output_fns = None
            response_type = type(response_t)
            if response_type == list:
                output_fns = " ".join(response_t)
            elif response_type == str:
                output_fns = response_t
            if output_fns is not None:
                quiet_print(f"report created: {output_fns} ", self.args)
            response[report_type] = response_t
            rtime = time() - stime
            quiet_print("finished in {0:.3f}s".format(rtime), self.args)
        return response

    def run_annotators_mp(self):
        from ..exceptions import SetupError
        import os
        import logging
        from ..base.mp_runners import init_worker, annot_from_queue
        from multiprocessing import Pool
        from ..system import get_max_num_concurrent_annotators_per_job

        if (
            self.args is None
            or self.manager is None
            or self.run_name is None
            or self.output_dir is None
        ):
            raise SetupError()
        num_workers = get_max_num_concurrent_annotators_per_job()
        if self.args.mp is not None:
            try:
                self.args.mp = int(self.args.mp)
                if self.args.mp >= 1:
                    num_workers = self.args.mp
            except:
                if self.logger:
                    self.logger.exception("error handling mp argument:")
        if self.logger:
            self.logger.info("num_workers: {}".format(num_workers))
        run_args = {}
        for module in self.run_annotators.values():
            inputpath = None
            # Make command
            if module.level == "variant":
                if module.conf.get("input_format"):
                    input_format = module.conf["input_format"]
                    if input_format == "crv":
                        inputpath = self.crvinput
                    elif input_format == "crx":
                        inputpath = self.crxinput
                    else:
                        raise Exception("Incorrect input_format value")
                        # inputpath = self.input
                else:
                    inputpath = self.crvinput
            elif module.level == "gene":
                inputpath = self.crginput
            # secondary_opts = []
            secondary_inputs = []
            if "secondary_inputs" in module.conf:
                secondary_module_names = module.conf["secondary_inputs"]
                for secondary_module_name in secondary_module_names:
                    secondary_module = self.annotators[secondary_module_name]
                    secondary_output_path = self.get_module_output_path(
                        secondary_module
                    )
                    if secondary_output_path is None:
                        if self.logger:
                            self.logger.warning(
                                f"secondary output file does not exist for {secondary_module_name}"
                            )
                    else:
                        secondary_inputs.append(
                            secondary_module.name.replace("=", r"\=")
                            + "="
                            + os.path.join(
                                self.output_dir, secondary_output_path
                            ).replace("=", r"\=")
                        )
            kwargs = {
                "script_path": module.script_path,
                "input_file": inputpath,
                "secondary_inputs": secondary_inputs,
                "quiet": self.args.quiet,
                "log_path": self.log_path,
                "run_conf": self.conf_run.get(module.name, {})
            }
            if self.run_name != None:
                kwargs["run_name"] = self.run_name
            if self.output_dir != None:
                kwargs["output_dir"] = self.output_dir
            run_args[module.name] = (module, kwargs)
        if self.logger and self.log_handler:
            self.logger.removeHandler(self.log_handler)
        start_queue = self.manager.Queue()
        end_queue = self.manager.Queue()
        all_mnames = set(self.run_annotators)
        assigned_mnames = set()
        done_mnames = set(self.done_annotators)
        queue_populated = self.manager.Value("c_bool", False)
        pool_args = [
            [start_queue, end_queue, queue_populated, self.status_writer]
        ] * num_workers
        with Pool(num_workers, init_worker) as pool:
            _ = pool.starmap_async(
                annot_from_queue,
                pool_args,
                error_callback=lambda _, mp_pool=pool: mp_pool.terminate(),
            )
            pool.close()
            for mname, module in self.run_annotators.items():
                if (
                    mname not in assigned_mnames
                    and set(module.secondary_module_names) <= done_mnames
                ):
                    start_queue.put(run_args[mname])
                    assigned_mnames.add(mname)
            while (
                assigned_mnames != all_mnames
            ):  # TODO not handling case where parent module errors out
                finished_module = end_queue.get()
                done_mnames.add(finished_module)
                for mname, module in self.run_annotators.items():
                    if (
                        mname not in assigned_mnames
                        and set(module.secondary_module_names) <= done_mnames
                    ):
                        start_queue.put(run_args[mname])
                        assigned_mnames.add(mname)
            queue_populated = True
            pool.join()
        self.log_path = os.path.join(self.output_dir, self.run_name + ".log")
        self.log_handler = logging.FileHandler(self.log_path, "a")
        formatter = logging.Formatter(
            "%(asctime)s %(name)-20s %(message)s", "%Y/%m/%d %H:%M:%S"
        )
        self.log_handler.setFormatter(formatter)
        if self.logger:
            self.logger.addHandler(self.log_handler)
        if len(self.run_annotators) > 0:
            self.annotator_ran = True

    def table_exists(self, cursor, table):
        sql = (
            'select name from sqlite_master where type="table" and '
            + 'name="'
            + table
            + '"'
        )
        cursor.execute(sql)
        if cursor.fetchone() == None:
            return False
        else:
            return True

    async def get_converter_format_from_crv(self):
        if self.run_name is None or self.output_dir is None:
            from ..exceptions import SetupError

            raise SetupError()
        import os

        converter_format = None
        fn = os.path.join(self.output_dir, self.run_name + ".crv")
        if os.path.exists(fn):
            f = open(fn)
            for line in f:
                if line.startswith("#input_format="):
                    converter_format = line.strip().split("=")[1]
                    break
            f.close()
        return converter_format

    async def get_mapper_info_from_crx(self):
        from ..exceptions import SetupError
        import os

        if self.run_name is None or self.output_dir is None:
            raise SetupError()
        title = None
        version = None
        modulename = None
        fn = os.path.join(self.output_dir, self.run_name + ".crx")
        if os.path.exists(fn):
            f = open(fn)
            for line in f:
                if line.startswith("#title="):
                    title = line.strip().split("=")[1]
                elif line.startswith("#version="):
                    version = line.strip().split("=")[1]
                elif line.startswith("#modulename="):
                    modulename = line.strip().split("=")[1]
                elif line.startswith("#") == False:
                    break
            f.close()
        return title, version, modulename

    async def write_job_info(self):
        from ..exceptions import SetupError
        from ..exceptions import NoInput
        import os
        import aiosqlite
        from datetime import datetime
        import json
        from ..module.local import get_local_module_info
        from ..exceptions import DatabaseError

        if (
            self.run_name is None
            or self.args is None
            or self.output_dir is None
        ):
            raise SetupError()
        if self.inputs is None:
            raise NoInput()
        dbpath = os.path.join(self.output_dir, self.run_name + ".sqlite")
        conn = await aiosqlite.connect(dbpath)
        cursor = await conn.cursor()
        if not self.append_mode:
            q = "drop table if exists info"
            await cursor.execute(q)
            q = "create table info (colkey text primary key, colval text)"
            await cursor.execute(q)
        modified = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        q = (
            'insert or replace into info values ("Result modified at", "'
            + modified
            + '")'
        )
        await cursor.execute(q)
        if not self.append_mode:
            created = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            q = "insert into info values (?, ?)"
            await cursor.execute(q, ("Result create at", created))
            q = "insert into info values (?, ?)"
            await cursor.execute(q, ("Input file name", ";".join(self.inputs)))
            genome_assembiles = ",".join(self.genome_assembiles or [])
            q = f"insert into info values (?, ?)"
            await cursor.execute(q, ("Input genome", genome_assembiles))
            q = "select count(*) from variant"
            await cursor.execute(q)
            r = await cursor.fetchone()
            if r is None:
                raise DatabaseError(msg="table variant does not exist.")
            no_input = str(r[0])
            q = (
                'insert into info values ("Number of unique input variants", "'
                + no_input
                + '")'
            )
            await cursor.execute(q)
            q = 'insert into info values ("oakvar", "{}")'.format(self.pkg_ver)
            await cursor.execute(q)
            q = 'insert into info values ("_converter_format", "{}")'.format(
                await self.get_converter_format_from_crv()
            )
            await cursor.execute(q)
            (
                mapper_title,
                mapper_version,
                mapper_modulename,
            ) = await self.get_mapper_info_from_crx()
            genemapper_str = "{} ({})".format(mapper_title, mapper_version)
            q = 'insert into info values ("Gene mapper", "{}")'.format(genemapper_str)
            await cursor.execute(q)
            q = 'insert into info values ("_mapper", "{}:{}")'.format(
                mapper_modulename, mapper_version
            )
            await cursor.execute(q)
            f = open(os.path.join(self.output_dir, self.run_name + ".crm"))
            for line in f:
                if line.startswith("#input_paths="):
                    input_path_dict_str = "=".join(line.strip().split("=")[1:]).replace(
                        '"', "'"
                    )
                    q = 'insert into info values ("_input_paths", "{}")'.format(
                        input_path_dict_str
                    )
                    await cursor.execute(q)
            q = f'insert into info values ("primary_transcript", "{",".join(self.args.primary_transcript)}")'
            await cursor.execute(q)
        q = 'select colval from info where colkey="annotators_desc"'
        await cursor.execute(q)
        r = await cursor.fetchone()
        if r is None:
            annotator_desc_dict = {}
        else:
            annotator_desc_dict = json.loads(r[0])
        q = "select name, displayname, version from variant_annotator"
        await cursor.execute(q)
        rows = list(await cursor.fetchall())
        q = "select name, displayname, version from gene_annotator"
        await cursor.execute(q)
        tmp_rows = list(await cursor.fetchall())
        if tmp_rows is not None:
            rows.extend(tmp_rows)
        annotators_str = ""
        annotator_version = {}
        annotators = []
        for row in rows:
            (name, displayname, version) = row
            if name in ["base", "tagsampler", "hg19", "hg18"]:
                continue
            if version is not None and version != "":
                annotators_str += "{} ({}), ".format(displayname, version)
                annotators.append("{}:{}".format(name, version))
            else:
                annotators_str += "{}, ".format(displayname)
                annotators.append("{}:".format(name))
            annotator_version[name] = version
            module_info = get_local_module_info(name)
            if module_info is not None and module_info.conf is not None:
                annotator_desc_dict[name] = module_info.conf.get("description", "")
        q = 'insert or replace into info values ("_annotator_desc", "{}")'.format(
            json.dumps(annotator_desc_dict).replace('"', "'")
        )
        await cursor.execute(q)
        if self.args.do_not_change_status != True:
            if self.status_writer:
                self.status_writer.queue_status_update(
                    "annotator_version", annotator_version, force=True
                )
        q = (
            'insert or replace into info values ("Annotators", "'
            + annotators_str
            + '")'
        )
        await cursor.execute(q)
        q = 'insert or replace into info values ("_annotators", "{}")'.format(
            ",".join(annotators)
        )
        await cursor.execute(q)
        await conn.commit()
        await cursor.close()
        await conn.close()

    def run_summarizers(self):
        from ..util.util import announce_module

        for module in self.ordered_summarizers:
            announce_module(module, status_writer=self.status_writer, args=self.args)
            self.run_summarizer(module)

    def run_summarizer(self, module):
        if self.args is None:
            from ..exceptions import SetupError

            raise SetupError()
        from ..util.util import load_class

        cmd = [module.script_path, "-l", "variant"]
        if self.run_name != None:
            cmd.extend(["-n", self.run_name])
        if self.output_dir != None:
            cmd.extend(["-d", self.output_dir])
        if self.verbose:
            if not self.args.quiet:
                print(" ".join(cmd))
        summarizer_cls = load_class(module.script_path, "")
        summarizer = summarizer_cls(cmd)
        summarizer.run()

    def clean_up_at_end(self):
        if self.output_dir is None or self.run_name is None:
            from ..exceptions import SetupError

            raise SetupError()
        import os

        fns = os.listdir(self.output_dir)
        for fn in fns:
            fn_path = os.path.join(self.output_dir, fn)
            if os.path.isfile(fn_path) == False:
                continue
            if fn.startswith(self.run_name):
                fn_end = fn.split(".")[-1]
                if fn_end in ["var", "gen", "crv", "crx", "crg", "crs", "crm", "crt"]:
                    os.remove(os.path.join(self.output_dir, fn))
                if fn.split(".")[-2:] == ["status", "json"]:
                    os.remove(os.path.join(self.output_dir, fn))

    async def write_admin_db(self, runtime, numinput):
        if self.args is None:
            from ..exceptions import SetupError

            raise SetupError()
        import os
        import aiosqlite
        from ..util.admin_util import get_admindb_path
        from ..util.util import quiet_print

        if runtime is None or numinput is None:
            return
        if os.path.exists(get_admindb_path()) == False:
            s = "{} does not exist.".format(get_admindb_path())
            if self.logger:
                self.logger.info(s)
            quiet_print(s, self.args)
            return
        db = await aiosqlite.connect(get_admindb_path())
        cursor = await db.cursor()
        q = 'update jobs set runtime={}, numinput={} where jobid="{}"'.format(
            runtime, numinput, self.args.jobid
        )
        await cursor.execute(q)
        await db.commit()
        await cursor.close()
        await db.close()

    def write_initial_status_json(self):
        if (
            self.run_name is None
            or self.inputs is None
            or self.args is None
            or self.output_dir is None
        ):
            from ..exceptions import SetupError

            raise SetupError()
        import os
        from datetime import datetime
        import json
        from ..util import admin_util as au

        status_fname = "{}.status.json".format(self.run_name)
        self.status_json_path = os.path.join(self.output_dir, status_fname)
        if os.path.exists(self.status_json_path) == True:
            with open(self.status_json_path) as f:
                try:
                    self.status_json = json.load(f)
                    self.pkg_ver = self.status_json.get("open_cravat_version")
                except:
                    self.pkg_ver = au.get_current_package_version()
            if self.status_json and self.status_json["status"] == "Submitted":
                self.status_json["job_dir"] = self.output_dir
                self.status_json["id"] = os.path.basename(
                    os.path.normpath(self.output_dir)
                )
                self.status_json["run_name"] = self.run_name
                self.status_json["db_path"] = os.path.join(
                    self.output_dir, self.run_name + ".sqlite"
                )
                self.status_json["orig_input_fname"] = [
                    os.path.basename(x) for x in self.inputs
                ]
                self.status_json["orig_input_path"] = self.inputs
                self.status_json["submission_time"] = datetime.now().isoformat()
                self.status_json["viewable"] = False
                self.status_json["note"] = self.args.note
                self.status_json["status"] = "Starting"
                self.status_json["reports"] = (
                    self.args.reports if self.args.reports != None else []
                )
                self.pkg_ver = au.get_current_package_version()
                self.status_json["open_cravat_version"] = self.pkg_ver
                annot_names = list(self.annotators.keys())
                annot_names.sort()
                if "original_input" in annot_names:
                    annot_names.remove("original_input")
                self.status_json["annotators"] = annot_names
                with open(self.status_json_path, "w") as wf:
                    wf.write(json.dumps(self.status_json, indent=2, sort_keys=True))
        else:
            self.status_json = {}
            self.status_json["job_dir"] = self.output_dir
            self.status_json["id"] = os.path.basename(os.path.normpath(self.output_dir))
            self.status_json["run_name"] = self.run_name
            self.status_json["db_path"] = os.path.join(
                self.output_dir, self.run_name + ".sqlite"
            )
            self.status_json["orig_input_fname"] = [
                os.path.basename(x) for x in self.inputs
            ]
            self.status_json["orig_input_path"] = self.inputs
            self.status_json["submission_time"] = datetime.now().isoformat()
            self.status_json["viewable"] = False
            self.status_json["note"] = self.args.note
            self.status_json["status"] = "Starting"
            self.status_json["reports"] = (
                self.args.reports if self.args.reports != None else []
            )
            self.pkg_ver = au.get_current_package_version()
            self.status_json["open_cravat_version"] = self.pkg_ver
            annot_names = list(self.annotators.keys())
            annot_names.sort()
            self.status_json["annotators"] = annot_names
            with open(self.status_json_path, "w") as wf:
                wf.write(json.dumps(self.status_json, indent=2, sort_keys=True))

    def write_smartfilters(self):
        if self.run_name is None or self.args is None or self.output_dir is None:
            from ..exceptions import SetupError

            raise SetupError()
        from time import time
        import os
        import json
        import sqlite3
        from ..util.util import filter_affected_cols
        from ..consts import base_smartfilters
        from ..util.util import quiet_print

        quiet_print("Indexing", self.args)
        dbpath = os.path.join(self.output_dir, self.run_name + ".sqlite")
        conn = sqlite3.connect(dbpath)
        cursor = conn.cursor()
        q = "create table if not exists smartfilters (name text primary key, definition text)"
        cursor.execute(q)
        ins_template = (
            "insert or replace into smartfilters (name, definition) values (?, ?);"
        )
        cols_to_index = set()
        for sf in base_smartfilters:
            cols_to_index |= filter_affected_cols(sf["filter"])
        if self.annotator_ran:
            for linfo in self.annotators.values():
                if linfo.smartfilters is not None:
                    for sf in linfo.smartfilters:
                        cols_to_index |= filter_affected_cols(sf["filter"])
                    mname = linfo.name
                    json_info = json.dumps(linfo.smartfilters)
                    cursor.execute(ins_template, (mname, json_info))
        cursor.execute("pragma table_info(variant)")
        variant_cols = {row[1] for row in cursor}
        cursor.execute("pragma table_info(gene)")
        gene_cols = {row[1] for row in cursor}
        cursor.execute('select name from sqlite_master where type="index"')
        existing_indices = {row[0] for row in cursor}
        for col in cols_to_index:
            if col in variant_cols:
                index_name = f"sf_variant_{col}"
                if index_name not in existing_indices:
                    q = f"create index if not exists {index_name} on variant ({col})"
                    quiet_print(f"\tvariant {col}", self.args)
                    st = time()
                    cursor.execute(q)
                    quiet_print(f"\tfinished in {time()-st:.3f}s", self.args)
            if col in gene_cols:
                index_name = f"sf_gene_{col}"
                if index_name not in existing_indices:
                    q = f"create index if not exists {index_name} on gene ({col})"
                    quiet_print(f"\tgene {col}", self.args)
                    st = time()
                    cursor.execute(q)
                    quiet_print(f"\tfinished in {time()-st:.3f}s", self.args)
        # Package filter
        if hasattr(self.args, "filter") and self.args.filter is not None:
            q = "create table if not exists viewersetup (datatype text, name text, viewersetup text, unique (datatype, name))"
            cursor.execute(q)
            filter_set = json.dumps({"filterSet": self.args.filter})
            q = f'insert or replace into viewersetup values ("filter", "quicksave-name-internal-use", \'{filter_set}\')'
            cursor.execute(q)
        conn.commit()
        cursor.close()
        conn.close()


class StatusWriter:
    def __init__(self, status_json_path):
        from time import time

        self.status_json_path = status_json_path
        self.status_queue = []
        self.status_json = None
        self.load_status_json()
        self.t = time()
        self.lock = False

    def load_status_json(self):
        import json

        f = open(self.status_json_path)
        lines = "\n".join(f.readlines())
        self.status_json = json.loads(lines)
        f.close()

    def queue_status_update(self, k, v, force=False):
        from time import time

        if self.status_json:
            self.status_json[k] = v
            tdif = time() - self.t
            if force == True or (tdif > 3 and self.lock == False):
                self.lock = True
                self.update_status_json()
                self.t = time()
                self.lock = False

    def update_status_json(self):
        import json

        with open(self.status_json_path, "w") as wf:
            json.dump(self.status_json, wf, indent=2, sort_keys=True)

    def get_status_json(self):
        return self.status_json

    def flush(self):
        from time import time

        self.lock = True
        self.update_status_json()
        self.t = time()
        self.lock = False


def add_parser_ov_run(subparsers):
    parser_ov_run = subparsers.add_parser(
        "run",
        prog="ov run input_file_path_1 input_file_path_2 ...",
        description="Run OakVar on input files.",
        help="Run a job",
        epilog="inputs should be the first argument",
    )
    parser_ov_run.r_return = "A string, a named list, or a dataframe. Output of reporters"  # type: ignore
    parser_ov_run.r_examples = [  # type: ignore
        "# Annotate the input file `input` with ClinVar and COSMIC modules ",
        "# and make a VCF-format report of annotated variants.",
        '#roakvar::run.input(inputs="input", annotators=list("clinvar", "cosmic"), reports="vcf")',
    ]
    parser_ov_run.add_argument(
        "inputs",
        nargs="*",
        default=[],
        help="Input file(s). One or more variant files in a supported format like VCF.  "
        + "See the -i/--input-format flag for supported formats. In the special case "
        + "where you want to add annotations to an existing OakVar analysis, "
        + "provide the output sqlite database from the previous run as input instead of a variant input file.",
    )
    parser_ov_run.add_argument(
        "-a",
        nargs="+",
        dest="annotators",
        default=[],
        help="Annotator module names or directories. If --package is used also, annotator modules defined with -a will be added.",
    )
    parser_ov_run.add_argument(
        "-A",
        nargs="+",
        dest="annotators_replace",
        default=[],
        help="Annotator module names or directories. If --package option also is used, annotator modules defined with -A will replace those defined with --package. -A has priority over -a.",
    )
    parser_ov_run.add_argument(
        "-e", nargs="+", dest="excludes", default=[], help="annotators to exclude"
    )
    parser_ov_run.add_argument("-n", dest="run_name", help="name of oakvar run")
    parser_ov_run.add_argument(
        "-d", dest="output_dir", default=None, help="directory for output files"
    )
    parser_ov_run.add_argument(
        "--startat",
        dest="startat",
        choices=[
            "converter",
            "mapper",
            "annotator",
            "aggregator",
            "postaggregator",
            "reporter",
        ],
        default=None,
        help="starts at given stage",
    )
    parser_ov_run.add_argument(
        "--endat",
        dest="endat",
        choices=[
            "converter",
            "mapper",
            "annotator",
            "aggregator",
            "postaggregator",
            "reporter",
        ],
        default=None,
        help="ends after given stage.",
    )
    parser_ov_run.add_argument(
        "--skip",
        dest="skip",
        nargs="+",
        choices=[
            "converter",
            "mapper",
            "annotator",
            "aggregator",
            "postaggregator",
            "reporter",
        ],
        default=None,
        help="skips given stage(s).",
    )
    parser_ov_run.add_argument(
        "-c", "--confpath", dest="confpath", default=None, help="path to a conf file"
    )
    parser_ov_run.add_argument(
        "--cs", dest="confs", default=None, help="configuration string"
    )
    parser_ov_run.add_argument(
        "-v", dest="verbose", action="store_true", default=None, help="verbose"
    )
    parser_ov_run.add_argument(
        "-t",
        nargs="+",
        dest="reports",
        default=[],
        help="Reporter types or reporter module directories",
    )
    parser_ov_run.add_argument(
        "-l",
        "--liftover",
        dest="genome",
        default=None,
        help="reference genome of input. OakVar will lift over to hg38 if needed.",
    )
    parser_ov_run.add_argument(
        "-x",
        dest="cleandb",
        action="store_true",
        help="deletes the existing result database and creates a new one.",
    )
    parser_ov_run.add_argument(
        "--newlog",
        dest="newlog",
        action="store_true",
        default=None,
        help="deletes the existing log file and creates a new one.",
    )
    parser_ov_run.add_argument(
        "--note",
        dest="note",
        default=None,
        help="note will be written to the run status file (.status.json)",
    )
    parser_ov_run.add_argument(
        "--mp",
        dest="mp",
        default=None,
        help="number of processes to use to run annotators",
    )
    parser_ov_run.add_argument(
        "-i",
        "--input-format",
        dest="forcedinputformat",
        default=None,
        help="Force input format",
    )
    parser_ov_run.add_argument(
        "--temp-files",
        dest="temp_files",
        action="store_true",
        default=None,
        help="Leave temporary files after run is complete.",
    )
    parser_ov_run.add_argument(
        "--keep-temp",
        dest="temp_files",
        action="store_true",
        default=None,
        help="Leave temporary files after run is complete.",
    )
    parser_ov_run.add_argument(
        "--writeadmindb",
        dest="writeadmindb",
        action="store_true",
        default=None,
        help="Write job information to admin db after job completion",
    )
    parser_ov_run.add_argument(
        "--jobid", dest="jobid", default=None, help="Job ID for server version"
    )
    parser_ov_run.add_argument(
        "--version",
        dest="show_version",
        action="store_true",
        default=None,
        help="Shows OakVar version.",
    )
    parser_ov_run.add_argument(
        "--separatesample",
        dest="separatesample",
        action="store_true",
        default=None,
        help="Separate variant results by sample",
    )
    parser_ov_run.add_argument(
        "--unique-variants",
        dest="unique_variants",
        action="store_true",
        default=None,
        help="Set to get only unique variants in output",
    )
    parser_ov_run.add_argument(
        "--primary-transcript",
        dest="primary_transcript",
        nargs="+",
        default=["mane"],
        help='"mane" for MANE transcripts as primary transcripts, or a path to a file of primary transcripts. MANE is default.',
    )
    parser_ov_run.add_argument(
        "--cleanrun",
        dest="clean",
        action="store_true",
        default=None,
        help="Deletes all previous output files for the job and generate new ones.",
    )
    parser_ov_run.add_argument(
        "--do-not-change-status",
        dest="do_not_change_status",
        action="store_true",
        default=None,
        help="Job status in status.json will not be changed",
    )
    parser_ov_run.add_argument(
        "--module-option",
        dest="module_option",
        nargs="*",
        help="Module-specific option in module_name.key=value syntax. For example, --module-option vcfreporter.type=separate",
    )
    parser_ov_run.add_argument(
        "--system-option",
        dest="system_option",
        nargs="*",
        default=[],
        help="System option in key=value syntax. For example, --system-option modules_dir=/home/user/oakvar/modules",
    )
    parser_ov_run.add_argument(
        "--quiet", action="store_true", default=None, help="Runs quietly."
    )
    parser_ov_run.add_argument(
        "--concise-report",
        dest="concise_report",
        action="store_true",
        default=None,
        help="Generate concise reports with default columns defined by each annotation module",
    )
    parser_ov_run.add_argument(
        "--package", dest="package", default=None, help="Use package"
    )
    parser_ov_run.add_argument("--filtersql", default=None, help="Filter SQL")
    parser_ov_run.add_argument(
        "--includesample", nargs="+", default=None, help="Sample IDs to include"
    )
    parser_ov_run.add_argument(
        "--excludesample", nargs="+", default=None, help="Sample IDs to exclude"
    )
    parser_ov_run.add_argument("--filter", default=None)
    parser_ov_run.add_argument(
        "-f", dest="filterpath", default=None, help="Path to a filter file"
    )
    parser_ov_run.add_argument(
        "--md",
        default=None,
        help="Specify the root directory of OakVar modules (annotators, etc)",
    )
    parser_ov_run.add_argument(
        "-m",
        dest="mapper_name",
        nargs="+",
        default=[],
        help="Mapper module name or mapper module directory",
    )
    parser_ov_run.add_argument(
        "-p",
        nargs="+",
        dest="postaggregators",
        default=[],
        help="Postaggregators to run. Additionally, tagsampler, casecontrol, varmeta, and vcfinfo will automatically run depending on conditions.",
    )
    parser_ov_run.add_argument(
        "--vcf2vcf",
        action="store_true",
        default=False,
        help="analyze with the vcf to vcf workflow. It is faster than a normal run, but only if both input and output formats are VCF.",
    )
    parser_ov_run.set_defaults(func=cli_run)



oc_store_url = "https://store.opencravat.org"
oc_publish_url = "https://publish.opencravat.org"
oc_store_module_url = oc_store_url + "/modules"
oc_manifest_url = oc_store_url + "/manifest.yml"
oc_manifest_fn = "oc_manifest.yml"
ov_store_email_key = "ov_store_email"
ov_store_pw_key = "ov_store_pw"
ov_store_cache_fn = "store.sqlite"
ov_store_last_updated_col = "last_updated"
ov_store_id_token_fname = ".token"
summary_table_cols = ["name", "store", "title", "type", "tags", "email", "publish_time"]
versions_table_cols = [
    "name",
    "store",
    "code_version",
    "data_version",
    "data_source",
    "code_size",
    "data_size",
    "publish_time",
]
logo_size = (180, 85)
ov_store_split_file_size = 1_000_000_000 # 1 GB parts
#ov_store_split_file_size = 100_000_000 # 100 MB parts
pack_ignore_fnames = [
    ".DS_Store",
    "__pycache__"
]
